import collections
import datetime
import json
import pathlib
import re
import subprocess
import tempfile
import traceback
import typing

import yaml

from filelock import FileLock
from sqlalchemy.orm import Session

from conda_store_server import action, api, conda_utils, orm, schema, utils
from conda_store_server.utils import BuildPathError


class LoggedStream:
    """Allows writing to storage via logging.StreamHandler"""

    def __init__(self, db, conda_store, build, prefix=None):
        self.db = db
        self.conda_store = conda_store
        self.build = build
        self.prefix = prefix

    def write(self, b, /):
        for line in b.split("\n"):
            # Skips empty lines
            if not line:
                continue
            if self.prefix is not None:
                line = self.prefix + line
            append_to_logs(self.db, self.conda_store, self.build, line + "\n")

    def flush(self):
        pass


def append_to_logs(db: Session, conda_store, build, logs: typing.Union[str, bytes]):
    # For instance, with local storage, this involves reading from and writing
    # to a file. Locking here prevents a race condition when multiple tasks
    # attempt to write to a shared resource, which is the log
    with FileLock(f"{build.build_path(conda_store)}.log.lock"):
        try:
            current_logs = conda_store.storage.get(build.log_key)
        except Exception:
            current_logs = b""

        if isinstance(logs, str):
            logs = logs.encode("utf-8")

        conda_store.storage.set(
            db,
            build.id,
            build.log_key,
            current_logs + logs,
            content_type="text/plain",
            artifact_type=schema.BuildArtifactType.LOGS,
        )


def set_build_started(db: Session, build: orm.Build):
    build.status = schema.BuildStatus.BUILDING
    build.started_on = datetime.datetime.utcnow()
    db.commit()


def set_build_failed(
    db: Session, build: orm.Build, status_info: typing.Optional[str] = None
):
    build.status = schema.BuildStatus.FAILED
    build.status_info = status_info
    build.ended_on = datetime.datetime.utcnow()
    db.commit()


def set_build_canceled(
    db: Session, build: orm.Build, status_info: typing.Optional[str] = None
):
    build.status = schema.BuildStatus.CANCELED
    build.status_info = status_info
    build.ended_on = datetime.datetime.utcnow()
    db.commit()


def set_build_completed(db: Session, conda_store, build: orm.Build):
    build.status = schema.BuildStatus.COMPLETED
    build.ended_on = datetime.datetime.utcnow()

    directory_build_artifact = orm.BuildArtifact(
        build_id=build.id,
        artifact_type=schema.BuildArtifactType.DIRECTORY,
        key=str(build.build_path(conda_store)),
    )
    db.add(directory_build_artifact)

    build.environment.current_build = build
    build.environment.specification = build.specification
    db.commit()


def build_cleanup(
    db: Session,
    conda_store,
    build_ids: typing.List[str] = None,
    reason: str = None,
    is_canceled: bool = False,
):
    """Walk through all builds in BUILDING state and check that they are actively running

    Build can get stuck in the building state due to worker
    spontaineously dying due to memory errors, killing container, etc.
    """
    status = "CANCELED" if is_canceled else "FAILED"
    reason = (
        reason
        or f"""
Build marked as {status} on cleanup due to being stuck in BUILDING state
and not present on workers. This happens for several reasons: build is
canceled, a worker crash from out of memory errors, worker was killed,
or error in conda-store
"""
    )

    inspect = conda_store.celery_app.control.inspect()
    active_tasks = inspect.active()
    if active_tasks is None:
        conda_store.log.warning(
            "build cleanup failed: celery broker does not support inspect"
        )
        return

    build_active_tasks = collections.defaultdict(list)
    for worker_name, tasks in active_tasks.items():
        for task in tasks:
            match = re.fullmatch("build-(\d+)-(.*)", str(task["id"]))
            if match:
                build_id, name = match.groups()
                build_active_tasks[build_id].append(task["name"])

    if build_ids:
        builds = [api.get_build(db, build_id) for build_id in build_ids]
    else:
        builds = api.list_builds(db, status=schema.BuildStatus.BUILDING)

    for build in builds:
        if (
            build.status == schema.BuildStatus.BUILDING
            and str(build.id) not in build_active_tasks
            and (
                build.started_on
                < (datetime.datetime.utcnow() - datetime.timedelta(seconds=5))
            )
        ):
            conda_store.log.warning(
                f"marking build {build.id} as {status} since stuck in BUILDING state and not present on workers"
            )
            append_to_logs(
                db,
                conda_store,
                build,
                reason,
            )
            if is_canceled:
                set_build_canceled(db, build)
            else:
                set_build_failed(db, build)


def build_conda_environment(db: Session, conda_store, build):
    """Build a conda environment with set uid/gid/and permissions and
    symlink the build to a named environment

    """
    try:
        set_build_started(db, build)
        # Note: even append_to_logs can fail due to filename size limit, so
        # check build_path length first
        conda_prefix = build.build_path(conda_store)
        append_to_logs(
            db,
            conda_store,
            build,
            f"starting build of conda environment {datetime.datetime.utcnow()} UTC\n",
        )

        settings = conda_store.get_settings(
            db=db,
            namespace=build.environment.namespace.name,
            environment_name=build.environment.name,
        )

        conda_prefix.parent.mkdir(parents=True, exist_ok=True)

        environment_prefix = build.environment_path(conda_store)
        if environment_prefix is not None:
            environment_prefix.parent.mkdir(parents=True, exist_ok=True)

        is_lockfile = build.specification.is_lockfile

        with utils.timer(conda_store.log, f"building conda_prefix={conda_prefix}"):
            if is_lockfile:
                context = action.action_save_lockfile(
                    specification=schema.LockfileSpecification.parse_obj(
                        build.specification.spec
                    ),
                    stdout=LoggedStream(
                        db=db,
                        conda_store=conda_store,
                        build=build,
                        prefix="action_save_lockfile: ",
                    ),
                )
            else:
                context = action.action_solve_lockfile(
                    settings.conda_command,
                    specification=schema.CondaSpecification.parse_obj(
                        build.specification.spec
                    ),
                    platforms=settings.conda_solve_platforms,
                    conda_flags=conda_store.conda_flags,
                    stdout=LoggedStream(
                        db=db,
                        conda_store=conda_store,
                        build=build,
                        prefix="action_solve_lockfile: ",
                    ),
                )

            conda_store.storage.set(
                db,
                build.id,
                build.conda_lock_key,
                json.dumps(
                    context.result, indent=4, cls=utils.CustomJSONEncoder
                ).encode("utf-8"),
                content_type="application/json",
                artifact_type=schema.BuildArtifactType.LOCKFILE,
            )

            conda_lock_spec = context.result

            context = action.action_fetch_and_extract_conda_packages(
                conda_lock_spec=conda_lock_spec,
                pkgs_dir=conda_utils.conda_root_package_dir(),
                stdout=LoggedStream(
                    db=db,
                    conda_store=conda_store,
                    build=build,
                    prefix="action_fetch_and_extract_conda_packages: ",
                ),
            )

            context = action.action_install_lockfile(
                conda_lock_spec=conda_lock_spec,
                conda_prefix=conda_prefix,
                stdout=LoggedStream(
                    db=db,
                    conda_store=conda_store,
                    build=build,
                    prefix="action_install_lockfile: ",
                ),
            )

        if environment_prefix is not None:
            utils.symlink(conda_prefix, environment_prefix)

        action.action_set_conda_prefix_permissions(
            conda_prefix=conda_prefix,
            permissions=settings.default_permissions,
            uid=settings.default_uid,
            gid=settings.default_gid,
            stdout=LoggedStream(
                db=db,
                conda_store=conda_store,
                build=build,
                prefix="action_set_conda_prefix_permissions: ",
            ),
        )

        action.action_add_conda_prefix_packages(
            db=db,
            conda_prefix=conda_prefix,
            build_id=build.id,
            stdout=LoggedStream(
                db=db,
                conda_store=conda_store,
                build=build,
                prefix="action_add_conda_prefix_packages: ",
            ),
        )

        context = action.action_get_conda_prefix_stats(
            conda_prefix,
            stdout=LoggedStream(
                db=db,
                conda_store=conda_store,
                build=build,
                prefix="action_get_conda_prefix_stats: ",
            ),
        )
        build.size = context.result["disk_usage"]

        set_build_completed(db, conda_store, build)
    # Always mark build as failed first since other functions may throw an
    # exception
    except subprocess.CalledProcessError as e:
        set_build_failed(db, build)
        conda_store.log.exception(e)
        append_to_logs(db, conda_store, build, e.output)
        raise e
    except BuildPathError as e:
        # Provide status_info, which will be exposed to the user, ONLY in this
        # case because the error message doesn't expose sensitive information
        set_build_failed(db, build, status_info=e.message)
        conda_store.log.exception(e)
        append_to_logs(db, conda_store, build, traceback.format_exc())
        raise e
    except Exception as e:
        set_build_failed(db, build)
        conda_store.log.exception(e)
        append_to_logs(db, conda_store, build, traceback.format_exc())
        raise e


def solve_conda_environment(db: Session, conda_store, solve: orm.Solve):
    settings = conda_store.get_settings(db=db)

    solve.started_on = datetime.datetime.utcnow()
    db.commit()

    context = action.action_solve_lockfile(
        conda_command=settings.conda_command,
        specification=schema.CondaSpecification.parse_obj(solve.specification.spec),
        platforms=[conda_utils.conda_platform()],
        conda_flags=conda_store.conda_flags,
    )
    conda_lock_spec = context.result

    action.action_add_lockfile_packages(
        db=db,
        conda_lock_spec=conda_lock_spec,
        solve_id=solve.id,
    )

    solve.ended_on = datetime.datetime.utcnow()
    db.commit()


def build_conda_env_export(db: Session, conda_store, build: orm.Build):
    conda_prefix = build.build_path(conda_store)
    settings = conda_store.get_settings(
        db=db,
        namespace=build.environment.namespace.name,
        environment_name=build.environment.name,
    )

    context = action.action_generate_conda_export(
        conda_command=settings.conda_command,
        conda_prefix=conda_prefix,
        stdout=LoggedStream(
            db=db,
            conda_store=conda_store,
            build=build,
            prefix="action_generate_conda_export: ",
        ),
    )

    conda_prefix_export = yaml.dump(context.result).encode("utf-8")

    conda_store.storage.set(
        db,
        build.id,
        build.conda_env_export_key,
        conda_prefix_export,
        content_type="text/yaml",
        artifact_type=schema.BuildArtifactType.YAML,
    )


def build_conda_pack(db: Session, conda_store, build: orm.Build):
    conda_prefix = build.build_path(conda_store)

    with utils.timer(
        conda_store.log, f"packaging archive of conda environment={conda_prefix}"
    ):
        with tempfile.TemporaryDirectory() as tmpdir:
            output_filename = pathlib.Path(tmpdir) / "environment.tar.gz"
            action.action_generate_conda_pack(
                conda_prefix=conda_prefix,
                output_filename=output_filename,
                stdout=LoggedStream(
                    db=db,
                    conda_store=conda_store,
                    build=build,
                    prefix="action_generate_conda_pack: ",
                ),
            )
            conda_store.storage.fset(
                db,
                build.id,
                build.conda_pack_key,
                output_filename,
                content_type="application/gzip",
                artifact_type=schema.BuildArtifactType.CONDA_PACK,
            )


def build_conda_docker(db: Session, conda_store, build: orm.Build):
    import warnings

    warnings.warn(
        "Generating Docker images is currently not supported, see "
        "https://github.com/conda-incubator/conda-store/issues/666"
    )
    return

    conda_prefix = build.build_path(conda_store)
    settings = conda_store.get_settings(
        db=db,
        namespace=build.environment.namespace.name,
        environment_name=build.environment.name,
    )

    try:
        with utils.timer(
            conda_store.log,
            f"packaging docker image of conda environment={conda_prefix}",
        ):
            context = action.action_generate_conda_docker(
                conda_prefix=conda_prefix,
                default_docker_image=utils.callable_or_value(
                    settings.default_docker_base_image, None
                ),
                container_registry=conda_store.container_registry,
                output_image_name=build.specification.name,
                output_image_tag=build.build_key,
            )
            append_to_logs(
                db,
                conda_store,
                build,
                "::group::action_generate_conda_docker\n"
                + context.stdout.getvalue()
                + "\n::endgroup::\n",
            )

            image = context.result

            if schema.BuildArtifactType.DOCKER_MANIFEST in settings.build_artifacts:
                conda_store.container_registry.store_image(
                    db, conda_store, build, image
                )

            if schema.BuildArtifactType.CONTAINER_REGISTRY in settings.build_artifacts:
                conda_store.container_registry.push_image(db, build, image)
    except Exception as e:
        conda_store.log.exception(e)
        append_to_logs(db, conda_store, build, traceback.format_exc())
        raise e


def build_constructor_installer(db: Session, conda_store, build: orm.Build):
    conda_prefix = build.build_path(conda_store)

    settings = conda_store.get_settings(
        db=db,
        namespace=build.environment.namespace.name,
        environment_name=build.environment.name,
    )

    with utils.timer(
        conda_store.log, f"creating installer for conda environment={conda_prefix}"
    ):
        with tempfile.TemporaryDirectory() as tmpdir:
            is_lockfile = build.specification.is_lockfile

            if is_lockfile:
                specification = schema.LockfileSpecification.parse_obj(
                    build.specification.spec
                )
            else:
                try:
                    # Tries to use the lockfile if it's available since it has
                    # pinned dependencies. This code is wrapped into try/except
                    # because the lockfile lookup might fail if the file is not
                    # in external storage or on disk, or if parsing fails
                    specification = schema.LockfileSpecification.parse_obj(
                        {
                            "name": build.specification.name,
                            "lockfile": json.loads(
                                conda_store.storage.get(build.conda_lock_key)
                            ),
                        }
                    )
                    is_lockfile = True
                except Exception as e:
                    conda_store.log.warning(
                        "Exception while obtaining lockfile, using specification",
                        exc_info=e,
                    )
                    specification = schema.CondaSpecification.parse_obj(
                        build.specification.spec
                    )

            context = action.action_generate_constructor_installer(
                conda_command=settings.conda_command,
                specification=specification,
                installer_dir=pathlib.Path(tmpdir),
                version=build.build_key,
                stdout=LoggedStream(
                    db=db,
                    conda_store=conda_store,
                    build=build,
                    prefix="action_generate_constructor_installer: ",
                ),
                is_lockfile=is_lockfile,
            )
            output_filename = context.result
            if output_filename is None:
                return
            conda_store.storage.fset(
                db,
                build.id,
                build.constructor_installer_key,
                output_filename,
                content_type="application/octet-stream",
                artifact_type=schema.BuildArtifactType.CONSTRUCTOR_INSTALLER,
            )



import contextlib
import hashlib
import json
import os
import pathlib
import re
import subprocess
import sys
import time

from filelock import FileLock


class CondaStoreError(Exception):
    @property
    def message(self):
        return self.args[0]


class BuildPathError(CondaStoreError):
    pass


def symlink(source, target):
    # Multiple builds call this, so this lock avoids race conditions on unlink
    # and symlink operations
    with FileLock(f"{target}.lock"):
        try:
            os.unlink(target)
        except FileNotFoundError:
            pass
        os.symlink(source, target)


def chmod(directory, permissions):
    if re.fullmatch("[0-7]{3}", permissions) is None:
        raise ValueError(
            f"chmod permissions={permissions} not 3 integer values between 0-7"
        )

    subprocess.check_output(["chmod", "-R", str(permissions), directory])


def chown(directory, uid, gid):
    if re.fullmatch(r"\d+", str(uid)) is None:
        raise ValueError(f"chown uid={uid} not integer value")

    if re.fullmatch(r"\d+", str(gid)) is None:
        raise ValueError(f"chown gid={gid} not integer value")

    subprocess.check_output(["chown", "-R", f"{uid}:{gid}", directory])


@contextlib.contextmanager
def chdir(directory: pathlib.Path):
    current_directory = os.getcwd()
    try:
        os.chdir(directory)
        yield
    finally:
        os.chdir(current_directory)


def du(path):
    """
    Pure Python equivalent of du -sb
    Based on https://stackoverflow.com/a/55648984/161801
    """
    if os.path.islink(path) or os.path.isfile(path):
        return os.lstat(path).st_size
    nbytes = 0
    seen = set()
    for dirpath, dirnames, filenames in os.walk(path):
        nbytes += os.lstat(dirpath).st_size
        for f in filenames:
            fp = os.path.join(dirpath, f)
            st = os.lstat(fp)
            if st.st_ino in seen:
                continue
            seen.add(st.st_ino)  # adds inode to seen list
            nbytes += st.st_size  # adds bytes to total
        for d in dirnames:
            dp = os.path.join(dirpath, d)
            if os.path.islink(dp):
                nbytes += os.lstat(dp).st_size
    return nbytes


def disk_usage(path: pathlib.Path):
    if sys.platform == "darwin":
        cmd = ["du", "-sAB1", str(path)]
    elif sys.platform == "linux":
        cmd = ["du", "-sb", str(path)]
    else:
        return str(du(path))

    output = subprocess.check_output(cmd, encoding="utf-8").split()[0]
    if sys.platform == "darwin":
        # mac du does not have the -b option to return bytes
        output = str(int(output) * 512)
    return output


@contextlib.contextmanager
def timer(logger, prefix):
    start_time = time.time()
    yield
    logger.info(f"{prefix} took {time.time() - start_time:.3f} [s]")


# conda-lock defines Channel.used_env_vars as frozenset. But frozenset is not
# serializable to JSON, so this creates a custom JSON encoder to handle this
# type
class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, frozenset):
            return list(obj)
        return super().default(obj)


def recursive_sort(v):
    """Recursively sort a nested python objects of lists, dicts,
    strings, ints, and floats

    """

    def sort_key(v):
        if isinstance(v, (list, dict, tuple)):
            return str(type(v))
        return f"{type(v)}-{v}"

    if isinstance(v, dict):
        sorted_keys = sorted(v.keys(), key=sort_key)
        return {k: recursive_sort(v[k]) for k in sorted_keys}
    elif isinstance(v, (tuple, list)):
        sorted_items = sorted(v, key=sort_key)
        return [recursive_sort(_) for _ in sorted_items]
    return v


def datastructure_hash(v):
    json_blob = json.dumps(recursive_sort(v), cls=CustomJSONEncoder)
    return hashlib.sha256(json_blob.encode("utf-8")).hexdigest()


def callable_or_value(v, *args, **kwargs):
    if callable(v):
        return v(*args, **kwargs)
    return v



"""Helper functions for user journeys."""

import json
import time
import uuid

from enum import Enum
from typing import Any, Callable, Dict, Optional, Union

import requests
import utils.time_utils as time_utils


TIMEOUT = 10


class BuildStatus(Enum):
    """Enum for API build status."""

    QUEUED = "QUEUED"
    BUILDING = "BUILDING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELED = "CANCELED"


class NamespaceStatus(Enum):
    OK = "ok"
    ERROR = "error"


class API:
    """
    Helper class for making requests to the API.
    These methods are used to build tests for user journeys
    """

    def __init__(
        self,
        base_url: str,
        token: str = "",
        username: str = "username",
        password: str = "password",
    ) -> None:
        self.base_url = base_url
        self.token = token
        if not token:
            # Log in if no token is provided to set the token
            self._login(username, password)

    def _make_request(
        self,
        endpoint: str,
        method: str = "GET",
        json_data: dict = None,
        headers: dict = None,
        timeout: int = TIMEOUT,
    ) -> requests.Response:
        """Make a request to the API."""
        url = f"{self.base_url}/{endpoint}"
        headers = headers or {}
        headers["Authorization"] = f"Bearer {self.token}"
        response = requests.request(
            method, url, json=json_data, headers=headers, timeout=timeout
        )
        response.raise_for_status()
        return response

    def get_logs(self, build_id: int) -> requests.Response:
        """Get the logs for the given build id.

        Parameters
        ----------
        build_id : int
            ID of the build to get the logs for

        Returns
        -------
        requests.Response
            Response from the conda-store-server. Logs are stored in the
            `.text` property, i.e. response.json()['text']
        """
        return self._make_request(f"api/v1/build/{build_id}/logs/")

    def _login(self, username: str, password: str) -> None:
        """Log in to the API and set an access token."""
        json_data = {"username": username, "password": password}
        response = requests.post(
            f"{self.base_url}/login", json=json_data, timeout=TIMEOUT
        )
        cookies = response.cookies.get_dict()
        token_response = requests.post(
            f"{self.base_url}/api/v1/token", cookies=cookies, timeout=TIMEOUT
        )
        data = token_response.json()
        self.token = data["data"]["token"]

    def create_namespace(self, namespace: Union[str, None] = None) -> requests.Response:
        """Create a namespace.

        Parameters
        ----------
        namespace : str
            Name of the namespace to create. If None, use a random namespace name

        Returns
        -------
        requests.Response
            Response from the conda-store server
        """
        if namespace is None:
            namespace = self.gen_random_namespace()

        self._make_request(f"api/v1/namespace/{namespace}", method="POST")
        return self._make_request(f"api/v1/namespace/{namespace}")

    def create_token(
        self, namespace: str, role: str, default_namespace: str = "default"
    ) -> requests.Response:
        """Create a token with a specified role in a specified namespace."""
        json_data = {
            "primary_namespace": default_namespace,
            "expiration": time_utils.get_iso8601_time(1),
            "role_bindings": {f"{namespace}/*": [role]},
        }
        return self._make_request("api/v1/token", method="POST", json_data=json_data)

    def create_environment(
        self,
        namespace: str,
        specification_path: str,
        max_iterations: int = 100,
        sleep_time: int = 5,
        wait: bool = True,
    ) -> requests.Response:
        """Create an environment.

        Parameters
        ----------
        namespace : str
            Namespace the environment should be written to
        specification_path : str
            Path to conda environment specification file
        max_iterations : int
            Max number of times to check whether the build completed before failing
        sleep_time : int
            Seconds to wait between each status check
        wait : bool
            If True, wait for the build to complete, fail, or be canceled before
            returning a response. If False, return the response from the specification
            POST immediately without waiting

        Returns
        -------
        requests.Response
            Response from the conda-store server's api/v1/build/{build_id}/
            endpoint
        """
        with open(specification_path, "r", encoding="utf-8") as file:
            specification_content = file.read()

        response = self._make_request(
            "api/v1/specification",
            method="POST",
            json_data={"namespace": namespace, "specification": specification_content},
        )
        if not wait:
            return response

        build_id = response.json()["data"]["build_id"]

        def check_status():
            status = self.get_build_status(build_id)
            if status in [BuildStatus.QUEUED, BuildStatus.BUILDING]:
                return False
            return status in [
                BuildStatus.FAILED,
                BuildStatus.CANCELED,
                BuildStatus.COMPLETED,
            ]

        wait_for_condition(check_status, timeout=240, interval=1)
        return self._make_request(f"api/v1/build/{build_id}/")

    def delete_environment(
        self, namespace: str, environment_name: str
    ) -> requests.Response:
        """Delete an environment."""
        return self._make_request(
            f"api/v1/environment/{namespace}/{environment_name}", method="DELETE"
        )

    def list_environments(self, namespace: Optional[str] = None):
        """List the environments in the given namespace.

        Parameters
        ----------
        namespace : Optional[str]
            Name of the namespace for which environments are to be retrieved.
            If None, all environments are retrieved.

        Returns
        -------
        requests.Response
            Response from the server containing the list of environments
        """
        if namespace:
            return self._make_request(
                f"api/v1/environment/?namespace={namespace}", method="GET"
            )
        return self._make_request("api/v1/environment/", method="GET")

    def delete_namespace(self, namespace: str) -> requests.Response:
        """Delete a namespace."""
        return self._make_request(f"api/v1/namespace/{namespace}", method="DELETE")

    @staticmethod
    def gen_random_namespace() -> str:
        """Generate a random namespace."""
        return uuid.uuid4().hex

    def set_active_build(
        self, namespace: str, environment: str, build_id: int
    ) -> requests.Response:
        """Set the active build for a given environment.

        Parameters
        ----------
        namespace : str
            Name of the namespace in which the environment lives
        environment : str
            Environment to set the build for
        build_id : int
            ID of the build to be activated for the given environment

        Returns
        -------
        requests.Response
            Response from the conda-store server.
        """
        return self._make_request(
            f"api/v1/environment/{namespace}/{environment}",
            method="PUT",
            json_data={"build_id": build_id},
        )

    def get_builds(
        self,
        environment: Optional[str] = None,
        namespace: Optional[str] = None,
    ) -> dict[str, Any]:
        """Get information about an environment.

        Parameters
        ----------
        namespace : Optional[str]
            Name of a namespace
        environment : Optional[str]
            Name of an environment

        Returns
        -------
        dict[str, Any]
            Dict of build properties; see API docs for
            api/v1/build/ for more information.
        """
        query_params = []
        if environment:
            query_params.append(f"name={environment}")

        if namespace:
            query_params.append(f"namespace={namespace}")

        return self._make_request(f'api/v1/build/?{"&".join(query_params)}').json()[
            "data"
        ]

    def get_environment(self, namespace: str, environment: str) -> dict[str, Any]:
        """Get information about an environment.

        Parameters
        ----------
        namespace : str
            Name of the namespace in which the environment lives
        environment : str
            Name of the environment

        Returns
        -------
        dict[str, Any]
            Dict of environment properties; see API docs for
            api/v1/environment/{namespace}/{environment}/ for more information.
        """
        return self._make_request(
            f"api/v1/environment/{namespace}/{environment}/"
        ).json()["data"]

    def cancel_build(self, build_id: int) -> requests.Response:
        """Cancel a build in progress.

        Parameters
        ----------
        build_id : int
            ID of the build to cancel

        Returns
        -------
        requests.Response
            Response from the server
        """
        return self._make_request(f"api/v1/build/{build_id}/cancel/", method="PUT")

    def get_build_status(self, build_id: int) -> BuildStatus:
        """Get the status of a build as a BuildStatus instance.

        Parameters
        ----------
        build_id : int
            ID of the build to get the status for

        Returns
        -------
        BuildStatus
            Build status for the given build ID
        """
        response = self._make_request(f"api/v1/build/{build_id}/")
        return BuildStatus(response.json()["data"]["status"])

    def get_lockfile(self, build_id: int) -> Dict[str, Any]:
        """Get a lockfile for the given build ID.
            Build for which the lockfile is to be retrieved

        Returns
        -------
        Dict[str, Any]
            Dictionary containing the lockfile version, metadata, and packages
        """
        return json.loads(self._make_request(f"api/v1/build/{build_id}/lockfile").text)


def wait_for_condition(
    condition: Callable[[], bool], timeout: int = 60, interval: int = 1
):
    """Call `condition` until it returns `True`.

    `condition` will be called every `interval` seconds up to a maximum of `timeout`
    seconds, at which point a ValueError is raised.

    Parameters
    ----------
    condition : Callable[[], bool]
        Function to call until True is returned
    timeout : int
        Number of seconds to continue calling `condition` for before timing out
    interval : int
        Number of seconds between consecutive calls
    """
    initial_time = time.time()
    while time.time() - initial_time < timeout:
        result = condition()
        if result:
            return

        time.sleep(interval)

    raise ValueError(
        f"Timeout after {timeout}s waiting for condition. Last result: {result}"
    )



import datetime

from typing import Any, Dict, List, Optional

import pydantic
import yaml

from fastapi import APIRouter, Body, Depends, HTTPException, Query, Request
from fastapi.responses import PlainTextResponse, RedirectResponse

from conda_store_server import __version__, api, orm, schema, utils
from conda_store_server.schema import Permissions
from conda_store_server.server import dependencies


router_api = APIRouter(
    tags=["api"],
    prefix="/api/v1",
)


def get_paginated_args(
    page: int = 1,
    order: Optional[str] = None,
    size: Optional[int] = None,
    sort_by: List[str] = Query([]),
    server=Depends(dependencies.get_server),
):
    if size is None:
        size = server.max_page_size
    size = min(size, server.max_page_size)
    offset = (page - 1) * size
    return {
        "limit": size,
        "offset": offset,
        "sort_by": sort_by,
        "order": order,
    }


def filter_distinct_on(
    query,
    distinct_on: List[str] = [],
    allowed_distinct_ons: Dict = {},
    default_distinct_on: List[str] = [],
):
    distinct_on = distinct_on or default_distinct_on
    distinct_on = [
        allowed_distinct_ons[d] for d in distinct_on if d in allowed_distinct_ons
    ]

    if distinct_on:
        return distinct_on, query.distinct(*distinct_on)
    return distinct_on, query


def get_sorts(
    order: str,
    sort_by: List[str] = [],
    allowed_sort_bys: Dict = {},
    required_sort_bys: List = [],
    default_sort_by: List = [],
    default_order: str = "asc",
):
    sort_by = sort_by or default_sort_by
    sort_by = [allowed_sort_bys[s] for s in sort_by if s in allowed_sort_bys]

    # required_sort_bys is needed when sorting is used with distinct
    # query see "SELECT DISTINCT ON expressions must match initial
    # ORDER BY expressions"
    if required_sort_bys != sort_by[: len(required_sort_bys)]:
        sort_by = required_sort_bys + sort_by

    if order not in {"asc", "desc"}:
        order = default_order

    order_mapping = {"asc": lambda c: c.asc(), "desc": lambda c: c.desc()}
    return [order_mapping[order](k) for k in sort_by]


def paginated_api_response(
    query,
    paginated_args,
    object_schema,
    sorts: List = [],
    exclude=None,
    allowed_sort_bys: Dict = {},
    required_sort_bys: List = [],
    default_sort_by: List = [],
    default_order: str = "asc",
):
    sorts = get_sorts(
        order=paginated_args["order"],
        sort_by=paginated_args["sort_by"],
        allowed_sort_bys=allowed_sort_bys,
        required_sort_bys=required_sort_bys,
        default_sort_by=default_sort_by,
        default_order=default_order,
    )

    count = query.count()
    query = (
        query.order_by(*sorts)
        .limit(paginated_args["limit"])
        .offset(paginated_args["offset"])
    )
    return {
        "status": "ok",
        "data": [object_schema.from_orm(_).dict(exclude=exclude) for _ in query.all()],
        "page": (paginated_args["offset"] // paginated_args["limit"]) + 1,
        "size": paginated_args["limit"],
        "count": count,
    }


@router_api.get(
    "/",
    response_model=schema.APIGetStatus,
)
async def api_status():
    return {"status": "ok", "data": {"version": __version__}}


@router_api.get(
    "/permission/",
    response_model=schema.APIGetPermission,
)
async def api_get_permissions(
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
):
    authenticated = entity is not None
    entity_binding_roles = auth.authorization.get_entity_bindings(entity)

    entity_binding_permissions = auth.authorization.get_entity_binding_permissions(
        entity
    )

    # convert Dict[str, set[enum]] -> Dict[str, List[str]]
    # to be json serializable
    entity_binding_permissions = {
        entity_arn: sorted([_.value for _ in entity_permissions])
        for entity_arn, entity_permissions in entity_binding_permissions.items()
    }

    return {
        "status": "ok",
        "data": {
            "authenticated": authenticated,
            "primary_namespace": (
                entity.primary_namespace
                if authenticated
                else conda_store.default_namespace
            ),
            "entity_permissions": entity_binding_permissions,
            "entity_roles": entity_binding_roles,
            "expiration": entity.exp if authenticated else None,
        },
    }


@router_api.get(
    "/usage/",
    response_model=schema.APIGetUsage,
)
async def api_get_usage(
    request: Request,
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        namespace_usage_metrics = auth.filter_namespaces(
            entity, api.get_namespace_metrics(db)
        )

        data = {}
        for namespace, num_environments, num_builds, storage in namespace_usage_metrics:
            data[namespace] = {
                "num_environments": num_environments,
                "num_builds": num_builds,
                "storage": storage,
            }

        return {
            "status": "ok",
            "data": data,
            "message": None,
        }


@router_api.post(
    "/token/",
    response_model=schema.APIPostToken,
)
async def api_post_token(
    request: Request,
    primary_namespace: Optional[str] = Body(None),
    expiration: Optional[datetime.datetime] = Body(None),
    role_bindings: Optional[Dict[str, List[str]]] = Body(None),
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
):
    if entity is None:
        entity = schema.AuthenticationToken(
            exp=datetime.datetime.now(tz=datetime.timezone.utc)
            + datetime.timedelta(days=1),
            primary_namespace=conda_store.default_namespace,
            role_bindings={},
        )

    new_entity = schema.AuthenticationToken(
        exp=expiration or entity.exp,
        primary_namespace=primary_namespace or entity.primary_namespace,
        role_bindings=role_bindings or auth.authorization.get_entity_bindings(entity),
    )

    if not auth.authorization.is_subset_entity_permissions(entity, new_entity):
        raise HTTPException(
            status_code=400,
            detail="Requested role_bindings are not a subset of current permissions",
        )

    if new_entity.exp > entity.exp:
        raise HTTPException(
            status_code=400,
            detail="Requested expiration of token is greater than current permissions",
        )

    return {
        "status": "ok",
        "data": {"token": auth.authentication.encrypt_token(new_entity)},
    }


@router_api.get(
    "/namespace/",
    response_model=schema.APIListNamespace,
    # don't send metadata_ and role_mappings
    response_model_exclude_defaults=True,
)
async def api_list_namespaces(
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
    paginated_args: Dict = Depends(get_paginated_args),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        orm_namespaces = auth.filter_namespaces(
            entity, api.list_namespaces(db, show_soft_deleted=False)
        )
        return paginated_api_response(
            orm_namespaces,
            paginated_args,
            schema.Namespace,
            exclude={"role_mappings", "metadata_"},
            allowed_sort_bys={
                "name": orm.Namespace.name,
            },
            default_sort_by=["name"],
        )


@router_api.get(
    "/namespace/{namespace}/",
    response_model=schema.APIGetNamespace,
)
async def api_get_namespace(
    namespace: str,
    request: Request,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request, namespace, {Permissions.NAMESPACE_READ}, require=True
        )

        namespace = api.get_namespace(db, namespace, show_soft_deleted=False)
        if namespace is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        return {
            "status": "ok",
            "data": schema.Namespace.from_orm(namespace).dict(),
        }


@router_api.post(
    "/namespace/{namespace}/",
    response_model=schema.APIAckResponse,
)
async def api_create_namespace(
    namespace: str,
    request: Request,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request, namespace, {Permissions.NAMESPACE_CREATE}, require=True
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm:
            raise HTTPException(status_code=409, detail="namespace already exists")

        try:
            api.create_namespace(db, namespace)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        db.commit()
        return {"status": "ok"}


@router_api.put(
    "/namespace/{namespace}/",
    response_model=schema.APIAckResponse,
)
async def api_update_namespace(
    namespace: str,
    request: Request,
    metadata: Dict[str, Any] = None,
    role_mappings: Dict[str, List[str]] = None,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            namespace,
            {
                Permissions.NAMESPACE_UPDATE,
                Permissions.NAMESPACE_ROLE_MAPPING_CREATE,
                Permissions.NAMESPACE_ROLE_MAPPING_DELETE,
            },
            require=True,
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        try:
            api.update_namespace(db, namespace, metadata, role_mappings)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        db.commit()
        return {"status": "ok"}


@router_api.put("/namespace/{namespace}/metadata", response_model=schema.APIAckResponse)
async def api_update_namespace_metadata(
    namespace: str,
    request: Request,
    metadata: Dict[str, Any] = None,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            namespace,
            {
                Permissions.NAMESPACE_UPDATE,
            },
            require=True,
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        try:
            api.update_namespace_metadata(db, namespace, metadata_=metadata)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        db.commit()
        return {"status": "ok"}


@router_api.get("/namespace/{namespace}/roles", response_model=schema.APIResponse)
async def api_get_namespace_roles(
    namespace: str,
    request: Request,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            namespace,
            {
                Permissions.NAMESPACE_READ,
                Permissions.NAMESPACE_ROLE_MAPPING_READ,
            },
            require=True,
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        try:
            data = api.get_namespace_roles(db, namespace)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        db.commit()
        return {
            "status": "ok",
            "data": [x.dict() for x in data],
        }


@router_api.delete("/namespace/{namespace}/roles", response_model=schema.APIAckResponse)
async def api_delete_namespace_roles(
    namespace: str,
    request: Request,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            namespace,
            {
                Permissions.NAMESPACE_READ,
                Permissions.NAMESPACE_ROLE_MAPPING_DELETE,
            },
            require=True,
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        try:
            api.delete_namespace_roles(db, namespace)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        db.commit()
        return {"status": "ok"}


@router_api.get("/namespace/{namespace}/role", response_model=schema.APIResponse)
async def api_get_namespace_role(
    namespace: str,
    request: Request,
    other_namespace: str,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            namespace,
            {
                Permissions.NAMESPACE_READ,
                Permissions.NAMESPACE_ROLE_MAPPING_READ,
            },
            require=True,
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        try:
            data = api.get_namespace_role(db, namespace, other=other_namespace)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        db.commit()
        if data is None:
            raise HTTPException(status_code=404, detail="failed to find role")
        return {
            "status": "ok",
            "data": data.dict(),
        }


@router_api.post("/namespace/{namespace}/role", response_model=schema.APIAckResponse)
async def api_create_namespace_role(
    namespace: str,
    request: Request,
    role_mapping: schema.APIPostNamespaceRole,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            namespace,
            {
                Permissions.NAMESPACE_READ,
                Permissions.NAMESPACE_ROLE_MAPPING_CREATE,
            },
            require=True,
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        try:
            api.create_namespace_role(
                db,
                namespace,
                other=role_mapping.other_namespace,
                role=role_mapping.role,
            )
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        db.commit()
        return {"status": "ok"}


@router_api.put("/namespace/{namespace}/role", response_model=schema.APIAckResponse)
async def api_update_namespace_role(
    namespace: str,
    request: Request,
    role_mapping: schema.APIPutNamespaceRole,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            namespace,
            {
                Permissions.NAMESPACE_READ,
                Permissions.NAMESPACE_ROLE_MAPPING_UPDATE,
            },
            require=True,
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        try:
            api.update_namespace_role(
                db,
                namespace,
                other=role_mapping.other_namespace,
                role=role_mapping.role,
            )
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        db.commit()
        return {"status": "ok"}


@router_api.delete("/namespace/{namespace}/role", response_model=schema.APIAckResponse)
async def api_delete_namespace_role(
    namespace: str,
    request: Request,
    role_mapping: schema.APIDeleteNamespaceRole,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            namespace,
            {
                Permissions.NAMESPACE_READ,
                Permissions.NAMESPACE_ROLE_MAPPING_DELETE,
            },
            require=True,
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        try:
            api.delete_namespace_role(db, namespace, other=role_mapping.other_namespace)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        db.commit()
        return {"status": "ok"}


@router_api.delete("/namespace/{namespace}/", response_model=schema.APIAckResponse)
async def api_delete_namespace(
    namespace: str,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request, namespace, {Permissions.NAMESPACE_DELETE}, require=True
        )

        namespace_orm = api.get_namespace(db, namespace)
        if namespace_orm is None:
            raise HTTPException(status_code=404, detail="namespace does not exist")

        try:
            conda_store.delete_namespace(db, namespace)
        except utils.CondaStoreError as e:
            raise HTTPException(status_code=400, detail=e.message)

        return {"status": "ok"}


@router_api.get(
    "/environment/",
    response_model=schema.APIListEnvironment,
)
async def api_list_environments(
    search: Optional[str] = None,
    namespace: Optional[str] = None,
    name: Optional[str] = None,
    status: Optional[schema.BuildStatus] = None,
    packages: Optional[List[str]] = Query([]),
    artifact: Optional[schema.BuildArtifactType] = None,
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
    paginated_args=Depends(get_paginated_args),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        orm_environments = auth.filter_environments(
            entity,
            api.list_environments(
                db,
                search=search,
                namespace=namespace,
                name=name,
                status=status,
                packages=packages,
                artifact=artifact,
                show_soft_deleted=False,
            ),
        )
        return paginated_api_response(
            orm_environments,
            paginated_args,
            schema.Environment,
            exclude={"current_build"},
            allowed_sort_bys={
                "namespace": orm.Namespace.name,
                "name": orm.Environment.name,
            },
            default_sort_by=["namespace", "name"],
        )


@router_api.get(
    "/environment/{namespace}/{environment_name}/",
    response_model=schema.APIGetEnvironment,
)
async def api_get_environment(
    namespace: str,
    environment_name: str,
    request: Request,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            f"{namespace}/{environment_name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )

        environment = api.get_environment(
            db, namespace=namespace, name=environment_name
        )
        if environment is None:
            raise HTTPException(status_code=404, detail="environment does not exist")

        return {
            "status": "ok",
            "data": schema.Environment.from_orm(environment).dict(
                exclude={"current_build"}
            ),
        }


@router_api.put(
    "/environment/{namespace}/{name}/",
    response_model=schema.APIAckResponse,
)
async def api_update_environment_build(
    namespace: str,
    name: str,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    build_id: int = Body(None, embed=True),
    description: str = Body(None, embed=True),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            f"{namespace}/{name}",
            {Permissions.ENVIRONMENT_UPDATE},
            require=True,
        )

        try:
            if build_id is not None:
                conda_store.update_environment_build(db, namespace, name, build_id)

            if description is not None:
                conda_store.update_environment_description(
                    db, namespace, name, description
                )

        except utils.CondaStoreError as e:
            raise HTTPException(status_code=400, detail=e.message)

        return {"status": "ok"}


@router_api.delete(
    "/environment/{namespace}/{name}/",
    response_model=schema.APIAckResponse,
)
async def api_delete_environment(
    namespace: str,
    name: str,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            f"{namespace}/{name}",
            {Permissions.ENVIRONMENT_DELETE},
            require=True,
        )

        try:
            conda_store.delete_environment(db, namespace, name)
        except utils.CondaStoreError as e:
            raise HTTPException(status_code=400, detail=e.message)

        return {"status": "ok"}


@router_api.get(
    "/specification/",
)
async def api_get_specification(
    request: Request,
    channel: List[str] = Query([]),
    conda: List[str] = Query([]),
    pip: List[str] = Query([]),
    format: schema.APIGetSpecificationFormat = Query(
        schema.APIGetSpecificationFormat.YAML
    ),
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
):
    with conda_store.get_db() as db:
        # GET is used for the solve to make this endpoint easily
        # cachable
        if pip:
            conda.append({"pip": pip})

        specification = schema.CondaSpecification(
            name="conda-store-solve",
            channels=channel,
            dependencies=conda,
        )

        try:
            task, solve_id = conda_store.register_solve(db, specification)
            task.wait()
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))

        solve = api.get_solve(db, solve_id)

        return {"solve": solve.packages}


@router_api.post(
    "/specification/",
    response_model=schema.APIPostSpecification,
)
async def api_post_specification(
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
    specification: str = Body(""),
    namespace: Optional[str] = Body(None),
    is_lockfile: Optional[bool] = Body(False, embed=True),
    environment_name: Optional[str] = Body("", embed=True),
    environment_description: Optional[str] = Body("", embed=True),
):
    with conda_store.get_db() as db:
        permissions = {Permissions.ENVIRONMENT_CREATE}

        default_namespace = (
            entity.primary_namespace if entity else conda_store.default_namespace
        )

        namespace_name = namespace or default_namespace
        namespace = api.get_namespace(db, namespace_name)
        if namespace is None:
            permissions.add(Permissions.NAMESPACE_CREATE)

        try:
            specification = yaml.safe_load(specification)
            if is_lockfile:
                lockfile_spec = {
                    "name": environment_name,
                    "description": environment_description,
                    "lockfile": specification,
                }
                specification = schema.LockfileSpecification.parse_obj(lockfile_spec)
            else:
                specification = schema.CondaSpecification.parse_obj(specification)
        except yaml.error.YAMLError:
            raise HTTPException(status_code=400, detail="Unable to parse. Invalid YAML")
        except utils.CondaStoreError as e:
            raise HTTPException(status_code=400, detail="\n".join(e.args[0]))
        except pydantic.ValidationError as e:
            raise HTTPException(status_code=400, detail=str(e))

        auth.authorize_request(
            request,
            f"{namespace_name}/{specification.name}",
            permissions,
            require=True,
        )

        try:
            build_id = conda_store.register_environment(
                db,
                specification,
                namespace_name,
                force=True,
                is_lockfile=is_lockfile,
            )
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))
        except utils.CondaStoreError as e:
            raise HTTPException(status_code=400, detail=str(e.message))

        return {"status": "ok", "data": {"build_id": build_id}}


@router_api.get("/build/", response_model=schema.APIListBuild)
async def api_list_builds(
    status: Optional[schema.BuildStatus] = None,
    packages: Optional[List[str]] = Query([]),
    artifact: Optional[schema.BuildArtifactType] = None,
    environment_id: Optional[int] = None,
    name: Optional[str] = None,
    namespace: Optional[str] = None,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
    paginated_args=Depends(get_paginated_args),
):
    with conda_store.get_db() as db:
        orm_builds = auth.filter_builds(
            entity,
            api.list_builds(
                db,
                status=status,
                packages=packages,
                artifact=artifact,
                environment_id=environment_id,
                name=name,
                namespace=namespace,
                show_soft_deleted=True,
            ),
        )
        return paginated_api_response(
            orm_builds,
            paginated_args,
            schema.Build,
            exclude={"specification", "packages", "build_artifacts"},
            allowed_sort_bys={
                "id": orm.Build.id,
                "started_on": orm.Build.started_on,
                "scheduled_on": orm.Build.scheduled_on,
                "ended_on": orm.Build.ended_on,
            },
            default_sort_by=["id"],
        )


@router_api.get("/build/{build_id}/", response_model=schema.APIGetBuild)
async def api_get_build(
    build_id: int,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        if build is None:
            raise HTTPException(status_code=404, detail="build id does not exist")

        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )

        return {
            "status": "ok",
            "data": schema.Build.from_orm(build).dict(exclude={"packages"}),
        }


@router_api.put(
    "/build/{build_id}/",
    response_model=schema.APIPostSpecification,
)
async def api_put_build(
    build_id: int,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        if build is None:
            raise HTTPException(status_code=404, detail="build id does not exist")

        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.ENVIRONMENT_UPDATE},
            require=True,
        )

        try:
            new_build = conda_store.create_build(
                db, build.environment_id, build.specification.sha256
            )
        except utils.CondaStoreError as e:
            raise HTTPException(status_code=400, detail=e.message)

        return {
            "status": "ok",
            "message": "rebuild triggered",
            "data": {"build_id": new_build.id},
        }


@router_api.put(
    "/build/{build_id}/cancel/",
    response_model=schema.APIAckResponse,
)
async def api_put_build_cancel(
    build_id: int,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        if build is None:
            raise HTTPException(status_code=404, detail="build id does not exist")

        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.BUILD_CANCEL},
            require=True,
        )

        if conda_store.celery_app.control.inspect().ping() is None:
            raise HTTPException(
                status_code=409,
                detail=(
                    "conda-store celery broker does not support task cancelation. "
                    "Use redis or rabbitmq message queues. "
                    "See docs for a more detailed explanation"
                ),
            )

        conda_store.celery_app.control.revoke(
            [
                f"build-{build_id}-conda-env-export",
                f"build-{build_id}-conda-pack",
                f"build-{build_id}-docker",
                f"build-{build_id}-constructor-installer",
                f"build-{build_id}-environment",
            ],
            terminate=True,
            signal="SIGTERM",
        )

        from conda_store_server.worker import tasks

        # Waits 5 seconds to ensure enough time for the task to actually be
        # canceled
        tasks.task_cleanup_builds.si(
            build_ids=[build_id],
            reason=f"""
    build {build_id} marked as CANCELED due to being canceled from the REST API
    """,
            is_canceled=True,
        ).apply_async(countdown=5)

        return {
            "status": "ok",
            "message": f"build {build_id} canceled",
        }


@router_api.delete(
    "/build/{build_id}/",
    response_model=schema.APIAckResponse,
)
async def api_delete_build(
    build_id: int,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        if build is None:
            raise HTTPException(status_code=404, detail="build id does not exist")

        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.BUILD_DELETE},
            require=True,
        )

        try:
            conda_store.delete_build(db, build_id)
        except utils.CondaStoreError as e:
            raise HTTPException(status_code=400, detail=e.message)

        return {"status": "ok"}


@router_api.get(
    "/build/{build_id}/packages/",
    response_model=schema.APIListCondaPackage,
)
async def api_get_build_packages(
    build_id: int,
    request: Request,
    search: Optional[str] = None,
    exact: Optional[str] = None,
    build: Optional[str] = None,
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
    paginated_args=Depends(get_paginated_args),
):
    with conda_store.get_db() as db:
        build_orm = api.get_build(db, build_id)
        if build_orm is None:
            raise HTTPException(status_code=404, detail="build id does not exist")

        auth.authorize_request(
            request,
            f"{build_orm.environment.namespace.name}/{build_orm.environment.name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )
        orm_packages = api.get_build_packages(
            db, build_orm.id, search=search, exact=exact, build=build
        )
        return paginated_api_response(
            orm_packages,
            paginated_args,
            schema.CondaPackage,
            allowed_sort_bys={
                "channel": orm.CondaChannel.name,
                "name": orm.CondaPackage.name,
            },
            default_sort_by=["channel", "name"],
            exclude={"channel": {"last_update"}},
        )


@router_api.get("/build/{build_id}/logs/")
async def api_get_build_logs(
    build_id: int,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        if build is None:
            raise HTTPException(status_code=404, detail="build id does not exist")

        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )

        return RedirectResponse(conda_store.storage.get_url(build.log_key))


@router_api.get(
    "/channel/",
    response_model=schema.APIListCondaChannel,
)
async def api_list_channels(
    conda_store=Depends(dependencies.get_conda_store),
    paginated_args=Depends(get_paginated_args),
):
    with conda_store.get_db() as db:
        orm_channels = api.list_conda_channels(db)
        return paginated_api_response(
            orm_channels,
            paginated_args,
            schema.CondaChannel,
            allowed_sort_bys={"name": orm.CondaChannel.name},
            default_sort_by=["name"],
        )


@router_api.get(
    "/package/",
    response_model=schema.APIListCondaPackage,
)
async def api_list_packages(
    search: Optional[str] = None,
    exact: Optional[str] = None,
    build: Optional[str] = None,
    paginated_args=Depends(get_paginated_args),
    conda_store=Depends(dependencies.get_conda_store),
    distinct_on: List[str] = Query([]),
):
    with conda_store.get_db() as db:
        orm_packages = api.list_conda_packages(
            db, search=search, exact=exact, build=build
        )
        required_sort_bys, distinct_orm_packages = filter_distinct_on(
            orm_packages,
            distinct_on=distinct_on,
            allowed_distinct_ons={
                "channel": orm.CondaChannel.name,
                "name": orm.CondaPackage.name,
                "version": orm.CondaPackage.version,
            },
        )
        return paginated_api_response(
            distinct_orm_packages,
            paginated_args,
            schema.CondaPackage,
            allowed_sort_bys={
                "channel": orm.CondaChannel.name,
                "name": orm.CondaPackage.name,
                "version": orm.CondaPackage.version,
            },
            default_sort_by=["channel", "name", "version", "build"],
            required_sort_bys=required_sort_bys,
            exclude={"channel": {"last_update"}},
        )


@router_api.get("/build/{build_id}/yaml/")
async def api_get_build_yaml(
    build_id: int,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        if build is None:
            raise HTTPException(status_code=404, detail="build id does not exist")

        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )
        return RedirectResponse(conda_store.storage.get_url(build.conda_env_export_key))


@router_api.get(
    "/environment/{namespace}/{environment_name}/conda-lock.yml",
    response_class=PlainTextResponse,
)
@router_api.get(
    "/environment/{namespace}/{environment_name}/conda-lock.yaml",
    response_class=PlainTextResponse,
)
@router_api.get(
    "/environment/{namespace}/{environment_name}/lockfile/",
    response_class=PlainTextResponse,
)
@router_api.get("/build/{build_id}/conda-lock.yml", response_class=PlainTextResponse)
@router_api.get(
    "/build/{build_id}/conda-lock.yaml",
    name="api_get_build_conda_lock_file",
    response_class=PlainTextResponse,
)
@router_api.get("/build/{build_id}/lockfile/", response_class=PlainTextResponse)
async def api_get_build_lockfile(
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    namespace: str = None,
    environment_name: str = None,
    build_id: int = None,
):
    with conda_store.get_db() as db:
        if build_id is None:
            environment = api.get_environment(
                db, namespace=namespace, name=environment_name
            )
            if environment is None:
                raise HTTPException(
                    status_code=404, detail="environment does not exist"
                )
            build = environment.current_build
        else:
            build = api.get_build(db, build_id)

        if build is None:
            raise HTTPException(status_code=404, detail="build id does not exist")

        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )

        build_artifacts = api.list_build_artifacts(
            db,
            build_id=build_id,
            included_artifact_types=[schema.BuildArtifactType.LOCKFILE],
        )
        # Checks if this is a legacy-style (v0.4.15) build, with a lockfile
        # generated by conda-store (newer builds use conda-lock)
        # https://github.com/conda-incubator/conda-store/issues/544
        if any(ba.key == "" for ba in build_artifacts):
            return api.get_build_lockfile_legacy(db, build_id)

        return RedirectResponse(conda_store.storage.get_url(build.conda_lock_key))


@router_api.get("/build/{build_id}/archive/")
async def api_get_build_archive(
    build_id: int,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )

        return RedirectResponse(conda_store.storage.get_url(build.conda_pack_key))


@router_api.get("/build/{build_id}/docker/")
async def api_get_build_docker_image_url(
    build_id: int,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    server=Depends(dependencies.get_server),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )

        if build.has_docker_manifest:
            url = f"{server.registry_external_url}/{build.environment.namespace.name}/{build.environment.name}:{build.build_key}"
            return PlainTextResponse(url)

        else:
            raise HTTPException(
                status_code=400,
                detail=f"Build {build_id} doesn't have a docker manifest",
            )


@router_api.get("/build/{build_id}/installer/")
async def api_get_build_installer(
    build_id: int,
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )

        if build.has_constructor_installer:
            return RedirectResponse(
                conda_store.storage.get_url(build.constructor_installer_key)
            )

        else:
            raise HTTPException(
                status_code=400,
                detail=f"Build {build_id} doesn't have an installer",
            )


@router_api.get(
    "/setting/",
    response_model=schema.APIGetSetting,
)
@router_api.get(
    "/setting/{namespace}/",
    response_model=schema.APIGetSetting,
)
@router_api.get(
    "/setting/{namespace}/{environment_name}/",
    response_model=schema.APIGetSetting,
)
async def api_get_settings(
    request: Request,
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    namespace: str = None,
    environment_name: str = None,
):
    with conda_store.get_db() as db:
        if namespace is None:
            arn = ""
        elif environment_name is None:
            arn = namespace
        else:
            arn = f"{namespace}/{environment_name}"

        auth.authorize_request(
            request,
            arn,
            {Permissions.SETTING_READ},
            require=True,
        )

        return {
            "status": "ok",
            "data": conda_store.get_settings(db, namespace, environment_name).dict(),
            "message": None,
        }


@router_api.put(
    "/setting/",
    response_model=schema.APIPutSetting,
)
@router_api.put(
    "/setting/{namespace}/",
    response_model=schema.APIPutSetting,
)
@router_api.put(
    "/setting/{namespace}/{environment_name}/",
    response_model=schema.APIPutSetting,
)
async def api_put_settings(
    request: Request,
    data: Dict[str, Any],
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    namespace: str = None,
    environment_name: str = None,
):
    with conda_store.get_db() as db:
        if namespace is None:
            arn = ""
        elif environment_name is None:
            arn = namespace
        else:
            arn = f"{namespace}/{environment_name}"

        auth.authorize_request(
            request,
            arn,
            {Permissions.SETTING_UPDATE},
            require=True,
        )

        try:
            conda_store.set_settings(db, namespace, environment_name, data)
        except ValueError as e:
            raise HTTPException(status_code=400, detail=str(e.args[0]))

        return {
            "status": "ok",
            "data": None,
            "message": f"global setting keys {list(data.keys())} updated",
        }



from typing import Optional

import yaml

from fastapi import APIRouter, Depends, Request
from fastapi.responses import RedirectResponse

from conda_store_server import api
from conda_store_server.action.generate_constructor_installer import (
    get_installer_platform,
)
from conda_store_server.schema import Permissions
from conda_store_server.server import dependencies


router_ui = APIRouter(tags=["ui"])


@router_ui.get("/create/")
async def ui_create_get_environment(
    request: Request,
    templates=Depends(dependencies.get_templates),
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
):
    with conda_store.get_db() as db:
        orm_namespaces = auth.filter_namespaces(
            entity, api.list_namespaces(db, show_soft_deleted=False)
        )

        default_namespace = (
            entity.primary_namespace if entity else conda_store.default_namespace
        )

        def sort_namespace(n):
            "Default namespace always first, then alphabetical"
            if n.name == default_namespace:
                return f"0{n.name}"
            return f"1{n.name}"

        context = {
            "request": request,
            "namespaces": sorted(orm_namespaces.all(), key=sort_namespace),
            "entity": entity,
        }

        return templates.TemplateResponse("create.html", context)


@router_ui.get("/")
async def ui_list_environments(
    request: Request,
    search: Optional[str] = None,
    templates=Depends(dependencies.get_templates),
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    server=Depends(dependencies.get_server),
    entity=Depends(dependencies.get_entity),
):
    with conda_store.get_db() as db:
        orm_environments = auth.filter_environments(
            entity,
            api.list_environments(db, search=search, show_soft_deleted=False),
        )

        context = {
            "request": request,
            "environments": orm_environments.all(),
            "registry_external_url": server.registry_external_url,
            "entity": entity,
            "platform": get_installer_platform(),
        }

        return templates.TemplateResponse("home.html", context)


@router_ui.get("/namespace/")
async def ui_list_namespaces(
    request: Request,
    templates=Depends(dependencies.get_templates),
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
):
    with conda_store.get_db() as db:
        orm_namespaces = auth.filter_namespaces(
            entity, api.list_namespaces(db, show_soft_deleted=False)
        )

        context = {
            "request": request,
            "namespaces": orm_namespaces.all(),
            "entity": entity,
        }

        return templates.TemplateResponse("namespace.html", context)


@router_ui.get("/environment/{namespace}/{environment_name}/")
async def ui_get_environment(
    namespace: str,
    environment_name: str,
    request: Request,
    templates=Depends(dependencies.get_templates),
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            f"{namespace}/{environment_name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )

        environment = api.get_environment(
            db, namespace=namespace, name=environment_name
        )
        if environment is None:
            return templates.TemplateResponse(
                "404.html",
                {
                    "request": request,
                    "message": f"environment namespace={namespace} name={environment_name} not found",
                },
                status_code=404,
            )

        specification = environment.current_build.specification
        is_lockfile = specification.is_lockfile
        spec = specification.spec
        description = spec["description"]
        if is_lockfile:
            spec = spec["lockfile"]

        context = {
            "request": request,
            "environment": environment,
            "entity": entity,
            "environment_description": description,
            "spec": yaml.dump(spec),
        }

        return templates.TemplateResponse("environment.html", context)


@router_ui.get("/environment/{namespace}/{environment_name}/edit/")
async def ui_edit_environment(
    namespace: str,
    environment_name: str,
    request: Request,
    templates=Depends(dependencies.get_templates),
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
):
    with conda_store.get_db() as db:
        auth.authorize_request(
            request,
            f"{namespace}/{environment_name}",
            {Permissions.ENVIRONMENT_CREATE},
            require=True,
        )

        environment = api.get_environment(
            db, namespace=namespace, name=environment_name
        )
        if environment is None:
            return templates.TemplateResponse(
                "404.html",
                {
                    "request": request,
                    "message": f"environment namespace={namespace} name={environment_name} not found",
                },
                status_code=404,
            )

        specification = environment.current_build.specification
        is_lockfile = specification.is_lockfile
        spec = specification.spec
        name = spec["name"]
        description = spec["description"]
        if is_lockfile:
            spec = spec["lockfile"]

        context = {
            "request": request,
            "environment": environment,
            "entity": entity,
            "environment_name": name,
            "environment_description": description,
            "specification": yaml.dump(spec),
            "is_lockfile": is_lockfile,
            "namespaces": [environment.namespace],
        }

        return templates.TemplateResponse("create.html", context)


@router_ui.get("/build/{build_id}/")
async def ui_get_build(
    build_id: int,
    request: Request,
    templates=Depends(dependencies.get_templates),
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    server=Depends(dependencies.get_server),
    entity=Depends(dependencies.get_entity),
):
    with conda_store.get_db() as db:
        build = api.get_build(db, build_id)
        if build is None:
            return templates.TemplateResponse(
                "404.html",
                {
                    "request": request,
                    "message": f"build id={build_id} not found",
                },
                status_code=404,
            )

        auth.authorize_request(
            request,
            f"{build.environment.namespace.name}/{build.environment.name}",
            {Permissions.ENVIRONMENT_READ},
            require=True,
        )

        context = {
            "request": request,
            "build": build,
            "registry_external_url": server.registry_external_url,
            "entity": entity,
            "spec": yaml.dump(build.specification.spec),
            "platform": get_installer_platform(),
        }

        return templates.TemplateResponse("build.html", context)


@router_ui.get("/user/")
async def ui_get_user(
    request: Request,
    templates=Depends(dependencies.get_templates),
    conda_store=Depends(dependencies.get_conda_store),
    auth=Depends(dependencies.get_auth),
    entity=Depends(dependencies.get_entity),
):
    with conda_store.get_db() as db:
        if entity is None:
            return RedirectResponse(request.url_for("get_login_method"))

        entity_binding_permissions = auth.authorization.get_entity_binding_permissions(
            entity
        )

        orm_namespaces = auth.filter_namespaces(
            entity, api.list_namespaces(db, show_soft_deleted=False)
        )

        system_metrics = api.get_system_metrics(db)

        namespace_usage_metrics = auth.filter_namespaces(
            entity, api.get_namespace_metrics(db)
        )

        context = {
            "request": request,
            "username": entity.primary_namespace,
            "namespaces": orm_namespaces,
            "entity_binding_permissions": entity_binding_permissions,
            "system_metrics": system_metrics,
            "namespace_usage_metrics": namespace_usage_metrics,
        }
        return templates.TemplateResponse("user.html", context)


@router_ui.get("/setting/")
@router_ui.get("/setting/{namespace}/")
@router_ui.get("/setting/{namespace}/{environment_name}/")
async def ui_get_setting(
    request: Request,
    templates=Depends(dependencies.get_templates),
    auth=Depends(dependencies.get_auth),
    conda_store=Depends(dependencies.get_conda_store),
    namespace: str = None,
    environment_name: str = None,
):
    with conda_store.get_db() as db:
        if namespace is None:
            arn = ""
        elif environment_name is None:
            arn = namespace
        else:
            arn = f"{namespace}/{environment_name}"

        auth.authorize_request(
            request,
            arn,
            {Permissions.SETTING_READ},
            require=True,
        )

        api_setting_url = str(request.url_for("api_put_settings"))
        if namespace is not None:
            api_setting_url += f"{namespace}/"
        if environment_name is not None:
            api_setting_url += f"{environment_name}/"

        context = {
            "request": request,
            "namespace": namespace,
            "environment_name": environment_name,
            "api_settings_url": api_setting_url,
            "settings": conda_store.get_settings(
                db, namespace=namespace, environment_name=environment_name
            ),
        }
    return templates.TemplateResponse("setting.html", context)



import pathlib
import stat

from conda_store_server import action, conda_utils, utils


@action.action
def action_set_conda_prefix_permissions(
    context,
    conda_prefix: pathlib.Path,
    permissions: str,
    uid: int,
    gid: int,
):
    conda_prefix = conda_prefix.resolve()

    # safeguard to try and prevent destructive actions
    if not conda_utils.is_conda_prefix(conda_prefix):
        raise ValueError("not a valid conda prefix")

    stat_info = conda_prefix.stat()

    if permissions is not None and oct(stat.S_IMODE(stat_info.st_mode))[-3:] != str(
        permissions
    ):
        context.log.info(
            f"modifying permissions of {conda_prefix} to permissions={permissions}"
        )
        with utils.timer(context.log, f"chmod of {conda_prefix}"):
            utils.chmod(conda_prefix, permissions)
    else:
        context.log.info(f"no changes for permissions of conda_prefix {conda_prefix}")

    if (
        uid is not None
        and gid is not None
        and (str(uid) != str(stat_info.st_uid) or str(gid) != str(stat_info.st_gid))
    ):
        context.log.info(
            f"modifying permissions of conda_prefix {conda_prefix} to uid={uid} and gid={gid}"
        )
        with utils.timer(context.log, f"chown of {conda_prefix}"):
            utils.chown(conda_prefix, uid, gid)
    else:
        context.log.info(f"no changes for gid and uid of conda_prefix {conda_prefix}")



import asyncio
import datetime
import pathlib
import re
import subprocess
import sys

import pytest
import yarl

from fastapi.responses import RedirectResponse
from traitlets import TraitError

from conda_store_server import (
    BuildKey,
    action,
    api,
    conda_utils,
    orm,
    schema,
    server,
    utils,
)
from conda_store_server.server.auth import DummyAuthentication


def test_action_decorator():
    @action.action
    def test_function(context):
        print("stdout")
        print("stderr", file=sys.stderr)
        if sys.platform == "win32":
            # echo is not a separate program on Windows
            context.run(["cmd", "/c", "echo subprocess"])
            context.run("echo subprocess_stdout", shell=True)
            context.run("echo subprocess_stderr>&2", shell=True)
            context.run(
                "echo subprocess_stderr_no_redirect>&2",
                shell=True,
                redirect_stderr=False,
            )
        else:
            context.run(["echo", "subprocess"])
            context.run("echo subprocess_stdout", shell=True)
            context.run("echo subprocess_stderr 1>&2", shell=True)
            context.run(
                "echo subprocess_stderr_no_redirect 1>&2",
                shell=True,
                redirect_stderr=False,
            )
        context.log.info("log")
        return pathlib.Path.cwd()

    context = test_function()
    assert (
        context.stdout.getvalue()
        == "stdout\nstderr\nsubprocess\nsubprocess_stdout\nsubprocess_stderr\nlog\n"
    )
    assert context.stderr.getvalue() == "subprocess_stderr_no_redirect\n"
    # test that action direction is not the same as outside function
    assert context.result != pathlib.Path.cwd()
    # test that temporary directory is cleaned up
    assert not context.result.exists()


@pytest.mark.parametrize(
    "specification",
    [
        "simple_specification",
        "simple_specification_with_pip",
    ],
)
def test_solve_lockfile(conda_store, specification, request):
    specification = request.getfixturevalue(specification)
    context = action.action_solve_lockfile(
        conda_command=conda_store.conda_command,
        specification=specification,
        platforms=[conda_utils.conda_platform()],
    )
    assert len(context.result["package"]) != 0


def test_solve_lockfile_valid_conda_flags(conda_store, simple_specification):
    context = action.action_solve_lockfile(
        conda_command=conda_store.conda_command,
        specification=simple_specification,
        platforms=[conda_utils.conda_platform()],
        conda_flags="--strict-channel-priority",
    )
    assert len(context.result["package"]) != 0


# Checks that conda_flags is used by conda-lock
def test_solve_lockfile_invalid_conda_flags(conda_store, simple_specification):
    with pytest.raises(
        Exception, match=(r"Command.*--this-is-invalid.*returned non-zero exit status")
    ):
        action.action_solve_lockfile(
            conda_command=conda_store.conda_command,
            specification=simple_specification,
            platforms=[conda_utils.conda_platform()],
            conda_flags="--this-is-invalid",
        )


@pytest.mark.parametrize(
    "specification",
    [
        "simple_specification",
        "simple_specification_with_pip",
    ],
)
def test_solve_lockfile_multiple_platforms(conda_store, specification, request):
    specification = request.getfixturevalue(specification)
    context = action.action_solve_lockfile(
        conda_command=conda_store.conda_command,
        specification=specification,
        platforms=["osx-64", "linux-64", "win-64", "osx-arm64"],
    )
    assert len(context.result["package"]) != 0


@pytest.mark.parametrize(
    "specification_name",
    [
        "simple_specification",
        "simple_specification_with_pip",
        "simple_lockfile_specification",
        "simple_lockfile_specification_with_pip",
    ],
)
def test_generate_constructor_installer(
    conda_store, specification_name, request, tmp_path
):
    specification = request.getfixturevalue(specification_name)
    installer_dir = tmp_path / "installer_dir"
    is_lockfile = specification_name in [
        "simple_lockfile_specification",
        "simple_lockfile_specification_with_pip",
    ]

    # Creates the installer
    context = action.action_generate_constructor_installer(
        conda_command=conda_store.conda_command,
        specification=specification,
        installer_dir=installer_dir,
        version="1",
        is_lockfile=is_lockfile,
    )

    # Checks that the installer was created
    installer = context.result
    assert installer.exists()

    tmp_dir = tmp_path / "tmp"

    # Runs the installer
    out_dir = pathlib.Path(tmp_dir) / "out"
    if sys.platform == "win32":
        subprocess.check_output([installer, "/S", f"/D={out_dir}"])
    else:
        subprocess.check_output([installer, "-b", "-p", str(out_dir)])

    # Checks the output directory
    assert out_dir.exists()
    lib_dir = out_dir / "lib"
    if specification_name in ["simple_specification", "simple_lockfile_specification"]:
        if sys.platform == "win32":
            assert any(str(x).endswith("zlib.dll") for x in out_dir.iterdir())
        elif sys.platform == "darwin":
            assert any(str(x).endswith("libz.dylib") for x in lib_dir.iterdir())
        else:
            assert any(str(x).endswith("libz.so") for x in lib_dir.iterdir())
    else:
        # Uses rglob to not depend on the version of the python
        # directory, which is where site-packages is located
        flask = pathlib.Path("site-packages") / "flask"
        assert any(str(x).endswith(str(flask)) for x in out_dir.rglob("*"))


def test_fetch_and_extract_conda_packages(tmp_path, simple_conda_lock):
    context = action.action_fetch_and_extract_conda_packages(
        conda_lock_spec=simple_conda_lock,
        pkgs_dir=tmp_path,
    )

    assert context.stdout.getvalue()


def test_install_specification(tmp_path, conda_store, simple_specification):
    conda_prefix = tmp_path / "test"

    action.action_install_specification(
        conda_command=conda_store.conda_command,
        specification=simple_specification,
        conda_prefix=conda_prefix,
    )

    assert conda_utils.is_conda_prefix(conda_prefix)


def test_install_lockfile(tmp_path, conda_store, simple_conda_lock):
    conda_prefix = tmp_path / "test"

    action.action_install_lockfile(
        conda_lock_spec=simple_conda_lock, conda_prefix=conda_prefix
    )

    assert conda_utils.is_conda_prefix(conda_prefix)


def test_generate_conda_export(conda_store, conda_prefix):
    context = action.action_generate_conda_export(
        conda_command=conda_store.conda_command, conda_prefix=conda_prefix
    )
    # The env name won't be correct because conda only sets the env name when
    # an environment is in an envs dir. See the discussion on PR #549.
    context.result["name"] = "test-prefix"

    schema.CondaSpecification.parse_obj(context.result)


def test_generate_conda_pack(tmp_path, conda_prefix):
    output_filename = tmp_path / "environment.tar.gz"

    action.action_generate_conda_pack(
        conda_prefix=conda_prefix,
        output_filename=output_filename,
    )

    assert output_filename.exists()


@pytest.mark.xfail(
    reason=(
        "Generating Docker images is currently not supported, see "
        "https://github.com/conda-incubator/conda-store/issues/666"
    )
)
def test_generate_conda_docker(conda_store, conda_prefix):
    action.action_generate_conda_docker(
        conda_prefix=conda_prefix,
        default_docker_image=utils.callable_or_value(
            conda_store.default_docker_base_image, None
        ),
        container_registry=conda_store.container_registry,
        output_image_name="test",
        output_image_tag="tag",
    )


def test_remove_not_conda_prefix(tmp_path):
    fake_conda_prefix = tmp_path / "test"
    fake_conda_prefix.mkdir()

    with pytest.raises(ValueError):
        action.action_remove_conda_prefix(fake_conda_prefix)


def test_remove_conda_prefix(tmp_path, simple_conda_lock):
    conda_prefix = tmp_path / "test"

    action.action_install_lockfile(
        conda_lock_spec=simple_conda_lock, conda_prefix=conda_prefix
    )

    assert conda_utils.is_conda_prefix(conda_prefix)

    action.action_remove_conda_prefix(conda_prefix)

    assert not conda_utils.is_conda_prefix(conda_prefix)
    assert not conda_prefix.exists()


@pytest.mark.skipif(
    sys.platform == "win32", reason="permissions are not supported on Windows"
)
def test_set_conda_prefix_permissions(tmp_path, conda_store, simple_conda_lock):
    conda_prefix = tmp_path / "test"

    action.action_install_lockfile(
        conda_lock_spec=simple_conda_lock, conda_prefix=conda_prefix
    )

    context = action.action_set_conda_prefix_permissions(
        conda_prefix=conda_prefix,
        permissions="755",
        uid=None,
        gid=None,
    )
    assert "no changes for permissions of conda_prefix" in context.stdout.getvalue()
    assert "no changes for gid and uid of conda_prefix" in context.stdout.getvalue()


def test_get_conda_prefix_stats(tmp_path, conda_store, simple_conda_lock):
    conda_prefix = tmp_path / "test"

    action.action_install_lockfile(
        conda_lock_spec=simple_conda_lock, conda_prefix=conda_prefix
    )

    context = action.action_get_conda_prefix_stats(conda_prefix)
    assert context.result["disk_usage"] > 0


def test_add_conda_prefix_packages(db, conda_store, simple_specification, conda_prefix):
    build_id = conda_store.register_environment(
        db, specification=simple_specification, namespace="pytest"
    )

    action.action_add_conda_prefix_packages(
        db=db,
        conda_prefix=conda_prefix,
        build_id=build_id,
    )

    build = api.get_build(db, build_id=build_id)
    assert len(build.package_builds) > 0


def test_add_lockfile_packages(
    db, conda_store, simple_specification, simple_conda_lock
):
    task, solve_id = conda_store.register_solve(db, specification=simple_specification)

    action.action_add_lockfile_packages(
        db=db,
        conda_lock_spec=simple_conda_lock,
        solve_id=solve_id,
    )

    solve = api.get_solve(db, solve_id=solve_id)
    assert len(solve.package_builds) > 0


@pytest.mark.parametrize(
    "is_legacy_build, build_key_version",
    [
        (False, 0),  # invalid
        (False, 1),  # long (legacy)
        (False, 2),  # shorter hash (default)
        (False, 3),  # hash-only (experimental)
        (True, 1),  # build_key_version doesn't matter because there's no lockfile
    ],
)
def test_api_get_build_lockfile(
    request,
    conda_store,
    db,
    simple_specification_with_pip,
    conda_prefix,
    is_legacy_build,
    build_key_version,
):
    # sets build_key_version
    if build_key_version == 0:  # invalid
        with pytest.raises(
            TraitError,
            match=(
                r"c.CondaStore.build_key_version: invalid build key version: 0, "
                r"expected: \(1, 2, 3\)"
            ),
        ):
            conda_store.build_key_version = build_key_version
        return  # invalid, nothing more to test
    conda_store.build_key_version = build_key_version
    assert BuildKey.current_version() == build_key_version
    assert BuildKey.versions() == (1, 2, 3)

    # initializes data needed to get the lockfile
    specification = simple_specification_with_pip
    specification.name = "this-is-a-long-environment-name"
    namespace = "pytest"

    class MyAuthentication(DummyAuthentication):
        # Skips auth (used in api_get_build_lockfile). Test version of request
        # has no state attr, which is returned in the real impl of this method.
        # So I have to overwrite the method itself.
        def authorize_request(self, *args, **kwargs):
            pass

    auth = MyAuthentication()
    build_id = conda_store.register_environment(
        db, specification=specification, namespace=namespace
    )
    db.commit()
    build = api.get_build(db, build_id=build_id)
    # makes this more visible in the lockfile
    build_id = 12345678
    build.id = build_id
    # makes sure the timestamp in build_key is always the same
    build.scheduled_on = datetime.datetime(2023, 11, 5, 3, 54, 10, 510258)
    environment = api.get_environment(db, namespace=namespace)

    # adds packages (returned in the lockfile)
    action.action_add_conda_prefix_packages(
        db=db,
        conda_prefix=conda_prefix,
        build_id=build_id,
    )

    key = "" if is_legacy_build else build.conda_lock_key

    # creates build artifacts
    build_artifact = orm.BuildArtifact(
        build_id=build_id,
        build=build,
        artifact_type=schema.BuildArtifactType.LOCKFILE,
        key=key,  # key value determines returned lockfile type
    )
    db.add(build_artifact)
    db.commit()

    # gets lockfile for this build
    res = asyncio.run(
        server.views.api.api_get_build_lockfile(
            request=request,
            conda_store=conda_store,
            auth=auth,
            namespace=namespace,
            environment_name=environment.name,
            build_id=build_id,
        )
    )

    if key == "":
        # legacy build: returns pinned package list
        lines = res.split("\n")
        assert len(lines) > 2
        assert lines[:2] == [
            f"#platform: {conda_utils.conda_platform()}",
            "@EXPLICIT",
        ]
        assert re.match("http.*//.*tar.bz2#.*", lines[2]) is not None
    else:
        # new build: redirects to lockfile generated by conda-lock
        def lockfile_url(build_key):
            return f"lockfile/{build_key}.yml"

        if build_key_version == 1:
            build_key = (
                "c7afdeffbe2bda7d16ca69beecc8bebeb29280a95d4f3ed92849e4047710923b-"
                "20231105-035410-510258-12345678-this-is-a-long-environment-name"
            )
        elif build_key_version == 2:
            build_key = "c7afdeff-1699156450-12345678-this-is-a-long-environment-name"
        elif build_key_version == 3:
            build_key = "c1f206a26263e1166e5b43548f69aa0c"
        else:
            raise ValueError(f"unexpected build_key_version: {build_key_version}")
        assert type(res) is RedirectResponse
        assert key == res.headers["location"]
        assert build.build_key == build_key
        assert BuildKey.get_build_key(build) == build_key
        assert build.parse_build_key(conda_store, build_key) == 12345678
        assert BuildKey.parse_build_key(conda_store, build_key) == 12345678
        assert lockfile_url(build_key) == build.conda_lock_key
        assert lockfile_url(build_key) == res.headers["location"]
        assert res.status_code == 307


def test_api_get_build_installer(
    request, conda_store, db, simple_specification_with_pip, conda_prefix
):
    # initializes data needed to get the installer
    specification = simple_specification_with_pip
    specification.name = "my-env"
    namespace = "pytest"

    class MyAuthentication(DummyAuthentication):
        # Skips auth (used in api_get_build_installer). Test version of request
        # has no state attr, which is returned in the real impl of this method.
        # So I have to overwrite the method itself.
        def authorize_request(self, *args, **kwargs):
            pass

    auth = MyAuthentication()
    build_id = conda_store.register_environment(
        db, specification=specification, namespace=namespace
    )
    db.commit()

    build = api.get_build(db, build_id=build_id)

    # creates build artifacts
    build_artifact = orm.BuildArtifact(
        build_id=build_id,
        build=build,
        artifact_type=schema.BuildArtifactType.CONSTRUCTOR_INSTALLER,
        key=build.constructor_installer_key,
    )
    db.add(build_artifact)
    db.commit()

    # gets installer for this build
    res = asyncio.run(
        server.views.api.api_get_build_installer(
            request=request,
            conda_store=conda_store,
            auth=auth,
            build_id=build_id,
        )
    )

    # redirects to installer
    def installer_url(build_key):
        ext = "exe" if sys.platform == "win32" else "sh"
        return f"installer/{build_key}.{ext}"

    assert type(res) is RedirectResponse
    assert build.constructor_installer_key == res.headers["location"]
    assert installer_url(build.build_key) == build.constructor_installer_key
    assert res.status_code == 307


def test_get_channel_url():
    conda_main = "https://conda.anaconda.org/main"
    repo_main = "https://repo.anaconda.com/pkgs/main"
    example = "https://example.com"

    assert conda_utils.get_channel_url(conda_main) == yarl.URL(repo_main)
    assert conda_utils.get_channel_url(f"{conda_main}/") == yarl.URL(repo_main)
    assert conda_utils.get_channel_url(example) == yarl.URL(example)



"""add status_info

Revision ID: d78e9889566a
Revises: b387747ca9b7
Create Date: 2023-11-07 12:25:04.416192

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "d78e9889566a"
down_revision = "b387747ca9b7"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("build", sa.Column("status_info", sa.UnicodeText(), nullable=True))


def downgrade():
    op.drop_column("build", "status_info")



"""adding a settings table

Revision ID: abd7248d5327
Revises: 16f65805dc8f
Create Date: 2023-05-11 16:38:12.210549

"""

import sqlalchemy as sa

from alembic import op


# revision identifiers, used by Alembic.
revision = "abd7248d5327"
down_revision = "16f65805dc8f"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "keyvaluestore",
        sa.Column("id", sa.Integer(), nullable=False, primary_key=True),
        sa.Column("prefix", sa.Unicode(), nullable=False, index=True),
        sa.Column("key", sa.Unicode(), nullable=False),
        sa.Column("value", sa.JSON(), nullable=True),
    )

    with op.batch_alter_table(
        "setting",
        table_args=[
            sa.UniqueConstraint("prefix", "key", name="_prefix_key_uc"),
        ],
    ):
        pass


def downgrade():
    op.drop_table("keyvaluestore")



import datetime
import enum
import functools
import os
import re
import sys

from typing import Any, Callable, Dict, List, Optional, Union

from conda_lock.lockfile.v1.models import Lockfile
from pydantic import BaseModel, Field, ValidationError, constr, validator
from pydantic.error_wrappers import ErrorWrapper

from conda_store_server import conda_utils, utils


def _datetime_factory(offset: datetime.timedelta):
    """utcnow datetime + timezone as string"""
    return datetime.datetime.utcnow() + offset


# namespace and name cannot contain "*" ":" "#" " " "/"
# this is a more restrictive list
ALLOWED_CHARACTERS = "A-Za-z0-9-+_@$&?^~.="
ARN_ALLOWED = f"^([{ALLOWED_CHARACTERS}*]+)/([{ALLOWED_CHARACTERS}*]+)$"


#########################
# Authentication Schema
#########################


class Permissions(enum.Enum):
    "Permissions map to conda-store actions"
    ENVIRONMENT_CREATE = "environment:create"
    ENVIRONMENT_READ = "environment::read"
    ENVIRONMENT_UPDATE = "environment::update"
    ENVIRONMENT_DELETE = "environment::delete"
    ENVIRONMENT_SOLVE = "environment::solve"
    BUILD_CANCEL = "build::cancel"
    BUILD_DELETE = "build::delete"
    NAMESPACE_CREATE = "namespace::create"
    NAMESPACE_READ = "namespace::read"
    NAMESPACE_UPDATE = "namespace::update"
    NAMESPACE_DELETE = "namespace::delete"
    NAMESPACE_ROLE_MAPPING_CREATE = "namespace-role-mapping::create"
    NAMESPACE_ROLE_MAPPING_READ = "namespace-role-mapping::read"
    NAMESPACE_ROLE_MAPPING_UPDATE = "namespace-role-mapping::update"
    NAMESPACE_ROLE_MAPPING_DELETE = "namespace-role-mapping::delete"
    SETTING_READ = "setting::read"
    SETTING_UPDATE = "setting::update"


class AuthenticationToken(BaseModel):
    exp: datetime.datetime = Field(
        default_factory=functools.partial(_datetime_factory, datetime.timedelta(days=1))
    )
    primary_namespace: str = "default"
    role_bindings: Dict[constr(regex=ARN_ALLOWED), List[str]] = {}


##########################
# Database Schema
##########################


class StorageBackend(enum.Enum):
    FILESYSTEM = "filesystem"
    S3 = "s3"


class CondaChannel(BaseModel):
    id: int
    name: str
    last_update: Optional[datetime.datetime]

    class Config:
        orm_mode = True


class CondaPackageBuild(BaseModel):
    id: int
    build: str
    sha256: str

    class Config:
        orm_mode = True


class CondaPackage(BaseModel):
    id: int
    channel: CondaChannel
    license: Optional[str]
    name: str
    version: str
    summary: Optional[str]

    class Config:
        orm_mode = True


class NamespaceRoleMapping(BaseModel):
    id: int
    entity: str
    role: str

    class Config:
        orm_mode = True


class NamespaceRoleMappingV2(BaseModel):
    id: int
    namespace: str
    other_namespace: str
    role: str

    class Config:
        orm_mode = True

    @classmethod
    def from_list(cls, lst):
        return cls(**{k: v for k, v in zip(cls.__fields__.keys(), lst)})


class Namespace(BaseModel):
    id: int
    name: constr(regex=f"^[{ALLOWED_CHARACTERS}]+$")  # noqa: F722
    metadata_: Dict[str, Any] = None
    role_mappings: List[NamespaceRoleMapping] = []

    class Config:
        orm_mode = True


class Specification(BaseModel):
    id: int
    name: str
    spec: dict
    sha256: str
    created_on: datetime.datetime

    class Config:
        orm_mode = True


class BuildArtifactType(enum.Enum):
    DIRECTORY = "DIRECTORY"
    LOCKFILE = "LOCKFILE"
    LOGS = "LOGS"
    YAML = "YAML"
    CONDA_PACK = "CONDA_PACK"
    DOCKER_BLOB = "DOCKER_BLOB"
    DOCKER_MANIFEST = "DOCKER_MANIFEST"
    CONTAINER_REGISTRY = "CONTAINER_REGISTRY"
    CONSTRUCTOR_INSTALLER = "CONSTRUCTOR_INSTALLER"


class BuildStatus(enum.Enum):
    QUEUED = "QUEUED"
    BUILDING = "BUILDING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELED = "CANCELED"


class BuildArtifact(BaseModel):
    id: int
    artifact_type: BuildArtifactType
    key: str

    class Config:
        orm_mode = True
        use_enum_values = True


class Build(BaseModel):
    id: int
    environment_id: int
    specification: Optional[Specification]
    packages: Optional[List[CondaPackage]]
    status: BuildStatus
    status_info: Optional[str]
    size: int
    scheduled_on: datetime.datetime
    started_on: Optional[datetime.datetime]
    ended_on: Optional[datetime.datetime]
    build_artifacts: Optional[List[BuildArtifact]]

    class Config:
        orm_mode = True
        use_enum_values = True


class Environment(BaseModel):
    id: int
    namespace: Namespace
    name: str
    current_build_id: int
    current_build: Optional[Build]

    description: Optional[str]

    class Config:
        orm_mode = True


class Settings(BaseModel):
    default_namespace: str = Field(
        "default",
        description="default namespace for conda-store",
        metadata={"global": True},
    )

    filesystem_namespace: str = Field(
        "filesystem",
        description="namespace to use for environments picked up via `CondaStoreWorker.watch_paths` on the filesystem",
        metadata={"global": True},
    )

    default_uid: Optional[int] = Field(
        None if sys.platform == "win32" else os.getuid(),
        description="default uid to assign to built environments",
        metadata={"global": True},
    )

    default_gid: Optional[int] = Field(
        None if sys.platform == "win32" else os.getgid(),
        description="default gid to assign to built environments",
        metadata={"global": True},
    )

    default_permissions: Optional[str] = Field(
        None if sys.platform == "win32" else "775",
        description="default file permissions to assign to built environments",
        metadata={"global": True},
    )

    storage_threshold: int = Field(
        5 * 1024**3,
        description="Storage threshold in bytes of minimum available storage required in order to perform builds",
        metadata={"global": True},
    )

    conda_command: str = Field(
        "mamba",
        description="conda executable to use for solves",
        metadata={"global": True},
    )

    conda_platforms: List[str] = Field(
        ["noarch", conda_utils.conda_platform()],
        description="Conda platforms to download package repodata.json from. By default includes current architecture and noarch",
        metadata={"global": True},
    )

    conda_max_solve_time: int = Field(
        5 * 60,  # 5 minute
        description="Maximum time in seconds to allow for solving a given conda environment",
        metadata={"global": True},
    )

    conda_indexed_channels: List[str] = Field(
        ["main", "conda-forge", "https://repo.anaconda.com/pkgs/main"],
        description="Conda channels to be indexed by conda-store at start.  Defaults to main and conda-forge.",
        metadata={"global": True},
    )

    build_artifacts_kept_on_deletion: List[BuildArtifactType] = Field(
        [
            BuildArtifactType.LOGS,
            BuildArtifactType.LOCKFILE,
            BuildArtifactType.YAML,
            # no possible way to delete these artifacts
            # in most container registries via api
            BuildArtifactType.CONTAINER_REGISTRY,
        ],
        description="artifacts to keep on build deletion",
        metadata={"global": True},
    )

    conda_solve_platforms: List[str] = Field(
        [conda_utils.conda_platform()],
        description="Conda platforms to solve environments for via conda-lock. Must include current platform.",
        metadata={"global": False},
    )

    conda_channel_alias: str = Field(
        "https://conda.anaconda.org",
        description="prepended url to associate with channel names",
        metadata={"global": False},
    )

    conda_default_channels: List[str] = Field(
        ["conda-forge"],
        description="channels that by default are included if channels are empty",
        metadata={"global": False},
    )

    conda_allowed_channels: List[str] = Field(
        [],
        description=(
            "Allowed conda channels to be used in conda environments. "
            "If set to empty list all channels are accepted (default). "
            "Example: "
            '["main", "conda-forge", "https://repo.anaconda.com/pkgs/main"]'
        ),
        metadata={"global": False},
    )

    conda_default_packages: List[str] = Field(
        [],
        description="Conda packages that included by default if none are included",
        metadata={"global": False},
    )

    conda_required_packages: List[str] = Field(
        [],
        description="Conda packages that are required to be within environment specification. Will raise a validation error is package not in specification",
        metadata={"global": False},
    )

    conda_included_packages: List[str] = Field(
        [],
        description="Conda packages that auto included within environment specification. Will not raise a validation error if package not in specification and will be auto added",
        metadata={"global": False},
    )

    pypi_default_packages: List[str] = Field(
        [],
        description="PyPi packages that included by default if none are included",
        metadata={"global": False},
    )

    pypi_required_packages: List[str] = Field(
        [],
        description="PyPi packages that are required to be within environment specification. Will raise a validation error is package not in specification",
        metadata={"global": False},
    )

    pypi_included_packages: List[str] = Field(
        [],
        description="PyPi packages that auto included within environment specification. Will not raise a validation error if package not in specification and will be auto added",
        metadata={"global": False},
    )

    build_artifacts: List[BuildArtifactType] = Field(
        [
            BuildArtifactType.LOCKFILE,
            BuildArtifactType.YAML,
            BuildArtifactType.CONDA_PACK,
            BuildArtifactType.CONSTRUCTOR_INSTALLER,
            *(
                [
                    BuildArtifactType.DOCKER_MANIFEST,
                    BuildArtifactType.CONTAINER_REGISTRY,
                ]
                if sys.platform == "linux"
                else []
            ),
        ],
        description="artifacts to build in conda-store. By default all of the artifacts",
        metadata={"global": False},
    )

    default_docker_base_image: Union[str, Callable] = Field(
        "registry-1.docker.io/library/debian:sid-slim",
        description="default base image used for the Dockerized environments. Make sure to have a proper glibc within image (highly discourage alpine/musl based images). Can also be callable function which takes the `orm.Build` object as input which has access to all attributes about the build such as install packages, requested packages, name, namespace, etc",
        metadata={"global": False},
    )

    @validator("build_artifacts", each_item=True)
    def check_build_artifacts(cls, v):
        return BuildArtifactType(v)

    @validator("build_artifacts_kept_on_deletion", each_item=True)
    def check_build_artifacts_kept_on_deletion(cls, v):
        return BuildArtifactType(v)

    class Config:
        use_enum_values = True


# Conda Environment
class CondaSpecificationPip(BaseModel):
    pip: List[str] = []

    @validator("pip", each_item=True)
    def check_pip(cls, v):
        from pkg_resources import Requirement

        allowed_pip_params = ["--index-url", "--extra-index-url", "--trusted-host"]

        if v.startswith("--"):
            match = re.fullmatch("(.+?)[ =](.*)", v)
            if match is None or match.group(1) not in allowed_pip_params:
                raise ValueError(
                    f"Invalid pip option '{v}' supported options are {allowed_pip_params}"
                )
        else:
            try:
                Requirement.parse(v)
            except Exception:
                raise ValueError(
                    f'Invalid pypi package dependency "{v}" ensure it follows peps https://peps.python.org/pep-0508/ and https://peps.python.org/pep-0440/'
                )

        return v


class CondaSpecification(BaseModel):
    name: constr(regex=f"^[{ALLOWED_CHARACTERS}]+$")  # noqa: F722
    channels: List[str] = []
    dependencies: List[Union[str, CondaSpecificationPip]] = []
    variables: Optional[Dict[str, Union[str, int]]]
    prefix: Optional[str]
    description: Optional[str] = ""

    @validator("dependencies", each_item=True)
    def check_dependencies(cls, v):
        from conda.models.match_spec import MatchSpec

        if not isinstance(v, str):
            return v  # ignore pip field

        try:
            MatchSpec(v)
        except Exception as e:
            print(e)
            raise ValueError(f"Invalid conda package dependency specification {v}")

        return v

    @classmethod
    def parse_obj(cls, specification):
        try:
            return super().parse_obj(specification)
        except ValidationError as e:
            # there can be multiple errors. Let's build a comprehensive summary
            # to return to the end user.

            # hr stands for "human readable"
            all_errors_hr = []

            for err in e.errors():
                error_type = err["type"]
                error_loc = err["loc"]

                # fallback case : if we can't figure out the error, let's build a default
                # one based on the data returned by Pydantic.
                human_readable_error = (
                    f"{err['msg']} (type={error_type}, loc={error_loc})"
                )

                if error_type == "type_error.none.not_allowed":
                    if error_loc[0] == "name":
                        human_readable_error = (
                            "The name of the environment cannot be empty."
                        )
                    else:
                        if len(error_loc) == 1:
                            human_readable_error = f"Invalid YAML : A forbidden `None` value has been encountered in section {error_loc[0]}"
                        elif len(error_loc) == 2:
                            human_readable_error = f"Invalid YAML : A forbidden `None` value has been encountered in section `{error_loc[0]}`, line {error_loc[1]}"

                all_errors_hr.append(human_readable_error)

            raise utils.CondaStoreError(all_errors_hr)


class LockfileSpecification(BaseModel):
    name: constr(regex=f"^[{ALLOWED_CHARACTERS}]+$")  # noqa: F722
    description: Optional[str] = ""
    lockfile: Lockfile

    @classmethod
    def parse_obj(cls, specification):
        # To show a human-readable error if no data is provided
        specification = {} if specification is None else specification
        # This uses pop because the version field must not be part of Lockfile
        # input. Otherwise, the input will be rejected. The version field is
        # hardcoded in the Lockfile schema and is only used when the output is
        # printed. So the code below validates that the version is 1 if present
        # and removes it to avoid the mentioned parsing error.
        lockfile = specification.get("lockfile")
        version = lockfile and lockfile.pop("version", None)
        if version not in (None, 1):
            # https://stackoverflow.com/questions/73968566/with-pydantic-how-can-i-create-my-own-validationerror-reason
            raise ValidationError(
                [
                    ErrorWrapper(
                        ValueError("expected no version field or version equal to 1"),
                        "lockfile -> version",
                    )
                ],
                LockfileSpecification,
            )

        return super().parse_obj(specification)

    def dict(self):
        res = super().dict()
        # The dict_for_output method includes the version field into the output
        # and excludes unset fields. Without the version field present,
        # conda-lock would reject a lockfile during parsing, so it wouldn't be
        # installable, that's why we need to include the version
        res["lockfile"] = self.lockfile.dict_for_output()
        return res

    def __str__(self):
        # This makes sure the format is suitable for output if this object is
        # converted to a string, which can also happen implicitly
        return str(self.dict())


###############################
#  Docker Registry Schema
###############################


def _docker_datetime_factory():
    """utcnow datetime + timezone as string"""
    return datetime.datetime.utcnow().astimezone().isoformat()


class DockerManifestLayer(BaseModel):
    mediaType: str = "application/vnd.docker.image.rootfs.diff.tar.gzip"
    size: int
    digest: str


class DockerManifestConfig(BaseModel):
    mediaType: str = "application/vnd.docker.container.image.v1+json"
    size: int
    digest: str


class DockerManifest(BaseModel):
    schemaVersion: int = 2
    mediaType: str = "application/vnd.docker.distribution.manifest.v2+json"
    config: DockerManifestConfig
    layers: List[DockerManifestLayer] = []


class DockerConfigConfig(BaseModel):
    Hostname: str = ""
    Domainname: str = ""
    User: str = ""
    AttachStdin: bool = False
    AttachStdout: bool = False
    AttachStderr: bool = False
    Tty: bool = False
    OpenStdin: bool = False
    StdinOnce: bool = False
    Env: List[str] = [
        "PATH=/opt/conda/bin:/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
    ]
    Cmd: List[str] = ["/bin/sh"]
    ArgsEscaped: bool = True
    Image: Optional[str] = None
    Volumes: Optional[List[str]]
    WorkingDir: str = ""
    Entrypoint: Optional[str] = None
    OnBuild: Optional[str] = None
    Labels: Optional[Dict[str, str]] = {"conda_store": "0.0.1"}


class DockerConfigRootFS(BaseModel):
    type: str = "layers"
    diff_ids: List[str] = []


class DockerConfigHistory(BaseModel):
    created: str = Field(default_factory=_docker_datetime_factory)
    created_by: str = ""


class DockerConfig(BaseModel):
    architecture: str = "amd64"
    os: str = "linux"
    config: DockerConfigConfig
    container: str
    container_config: DockerConfigConfig
    created: str = Field(default_factory=_docker_datetime_factory)
    docker_version: str = "18.09.7"
    history: List[DockerConfigHistory] = []
    rootfs: DockerConfigRootFS


# https://docs.docker.com/registry/spec/api/#errors-2
class DockerRegistryError(enum.Enum):
    NAME_UNKNOWN = {
        "message": "repository name not known to registry",
        "detail": "This is returned if the name used during an operation is unknown to the registry",
        "status": 404,
    }
    BLOB_UNKNOWN = {
        "message": "blob unknown to registry",
        "detail": "This error may be returned when a blob is unknown to the registry in a specified repository. This can be returned with a standard get or if a manifest references an unknown layer during upload",
        "status": 404,
    }
    MANIFEST_UNKNOWN = {
        "message": "manifest unknown",
        "detail": "This error is returned when the manifest, identified by name and tag is unknown to the repository",
        "status": 404,
    }
    UNAUTHORIZED = {
        "message": "authentication required",
        "detail": "The access controller was unable to authenticate the client. Often this will be accompanied by a Www-Authenticate HTTP response header indicating how to authenticate",
        "status": 401,
    }
    UNSUPPORTED = {
        "message": "The operation is unsupported",
        "detail": "The operation was unsupported due to a missing implementation or invalid set of parameters",
        "status": 405,
    }
    DENIED = {
        "message": "requested access to the resource is denied",
        "detail": "The access controller denied access for the operation on a resource",
        "status": 403,
    }


# API Response Objects
class APIStatus(enum.Enum):
    OK = "ok"
    ERROR = "error"


class APIResponse(BaseModel):
    status: APIStatus
    data: Optional[Any]
    message: Optional[str]


class APIPaginatedResponse(APIResponse):
    page: int
    size: int
    count: int


class APIAckResponse(BaseModel):
    status: APIStatus
    message: Optional[str]


# GET /api/v1
class APIGetStatusData(BaseModel):
    version: str


class APIGetStatus(APIResponse):
    data: APIGetStatusData


# GET /api/v1/permission
class APIGetPermissionData(BaseModel):
    authenticated: bool
    primary_namespace: str
    entity_permissions: Dict[str, List[str]]
    entity_roles: Dict[str, List[str]]
    expiration: Optional[datetime.datetime]


class APIGetPermission(APIResponse):
    data: APIGetPermissionData


# POST /api/v1/token
class APIPostTokenData(BaseModel):
    token: str


class APIPostToken(APIResponse):
    data: APIPostTokenData


# GET /api/v1/namespace
class APIListNamespace(APIPaginatedResponse):
    data: List[Namespace]


# GET /api/v1/namespace/{name}
class APIGetNamespace(APIResponse):
    data: Namespace


# POST /api/v1/namespace/{name}/role
class APIPostNamespaceRole(BaseModel):
    other_namespace: str
    role: str


# PUT /api/v1/namespace/{name}/role
class APIPutNamespaceRole(BaseModel):
    other_namespace: str
    role: str


# DELETE /api/v1/namespace/{name}/role
class APIDeleteNamespaceRole(BaseModel):
    other_namespace: str


# GET /api/v1/environment
class APIListEnvironment(APIPaginatedResponse):
    data: List[Environment]


# GET /api/v1/environment/{namespace}/{name}
class APIGetEnvironment(APIResponse):
    data: Environment


# GET /api/v1/specification
class APIGetSpecificationFormat(enum.Enum):
    YAML = "yaml"
    LOCKFILE = "lockfile"


# POST /api/v1/specification
class APIPostSpecificationData(BaseModel):
    build_id: int


class APIPostSpecification(APIResponse):
    data: APIPostSpecificationData


# GET /api/v1/build
class APIListBuild(APIPaginatedResponse):
    data: List[Build]


# GET /api/v1/build/1
class APIGetBuild(APIResponse):
    data: Build


# GET /api/v1/channel
class APIListCondaChannel(APIPaginatedResponse):
    data: List[CondaChannel]


# GET /api/v1/package
class APIListCondaPackage(APIPaginatedResponse):
    data: List[CondaPackage]


# GET /api/v1/setting/*/*
class APIGetSetting(APIResponse):
    data: Dict[str, Any]


# PUT /api/v1/setting/*/*
class APIPutSetting(APIResponse):
    pass


# GET /api/v1/usage/
class APIGetUsage(APIResponse):
    data: Dict[str, Dict[str, Any]]



from conda_store_server.utils import disk_usage, du


# TODO: Add tests for the other functions in utils.py


def test_disk_usage(tmp_path):
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()

    # This varies across OSes
    dir_size = du(test_dir)
    assert abs(dir_size - int(disk_usage(test_dir))) <= 1000

    test_file = test_dir / "test_file"
    test_file.write_text("a" * 1000)
    test_file2 = test_dir / "test_file2"
    test_file2.write_text("b" * 1000)
    # Test hard links
    test_file_hardlink = test_dir / "test_file_hardlink"
    test_file_hardlink.hardlink_to(test_file)

    # The exact values depend on the block size and other details. Just check
    # that it is in the right range. Note that disk_usage uses the du command
    # on Mac and Linux but uses the Python du() function on Windows.
    val = disk_usage(test_file)
    assert isinstance(val, str)
    assert 1000 <= int(val) <= 1500
    assert 1000 <= du(test_file) <= 1500

    val = disk_usage(test_dir)
    assert isinstance(val, str)
    assert 2000 + dir_size <= int(val) <= 2700 + dir_size
    assert 2000 + dir_size <= du(test_dir) <= 2700 + dir_size



from conda_store_server import api, schema, testing


def test_testing_initialize_database(db, conda_store):
    config = {
        "namespace1": {
            "name1": schema.CondaSpecification(
                name="name1",
                channels=["conda-forge"],
                dependencies=["numpy"],
            ),
            "name2": schema.CondaSpecification(
                name="name2",
                channels=["defaults"],
                dependencies=["flask"],
            ),
        },
        "namespace2": {
            "name3": schema.CondaSpecification(
                name="name3",
                channels=["bioconda"],
                dependencies=["numba"],
            )
        },
    }

    testing.seed_conda_store(db, conda_store, config)

    assert len(api.list_namespaces(db).all()) == 2
    assert len(api.list_environments(db).all()) == 3
    assert len(api.list_builds(db).all()) == 3
    assert len(api.list_solves(db).all()) == 3
    assert len(api.list_conda_packages(db).all()) == 3



"""User journey tests for the API."""

import os

import pytest
import utils.api_utils as utils


@pytest.fixture(scope="session")
def base_url() -> str:
    """Get the base URL for the API."""
    base = os.getenv("CONDA_STORE_BASE_URL", "http://localhost:8080")
    return f"{base}/conda-store"


@pytest.fixture(scope="session")
def token(base_url) -> str:
    """Get the token for the API."""
    return os.getenv("CONDA_STORE_TOKEN", "")


@pytest.mark.user_journey
@pytest.mark.parametrize(
    "specification_path",
    [
        ("tests/user_journeys/test_data/simple_environment.yaml"),
    ],
)
def test_admin_user_can_create_environment(
    base_url: str, token: str, specification_path: str
) -> None:
    """Test that an admin user can create an environment."""
    api = utils.API(base_url=base_url, token=token)
    namespace = "default"
    environment = api.create_environment(namespace, specification_path).json()["data"][
        "specification"
    ]["name"]
    api.delete_environment(namespace, environment)


@pytest.mark.user_journey
@pytest.mark.parametrize(
    "specification_path",
    [
        ("tests/user_journeys/test_data/simple_environment.yaml"),
    ],
)
def test_admin_login_and_delete_shared_environment(
    base_url: str, specification_path: str
) -> None:
    """Test that an admin can login and create/delete an env in a shared namespace."""
    api = utils.API(base_url=base_url)

    # Create a shared namespace; default permissions for namepace/environment
    # */* is admin
    namespace = api.create_namespace().json()["data"]["name"]
    environment = api.create_environment(
        namespace,
        specification_path,
    ).json()["data"][
        "specification"
    ]["name"]

    api.delete_environment(namespace, environment)
    api.delete_namespace(namespace)


@pytest.mark.user_journey
@pytest.mark.parametrize(
    "specification_path",
    [
        ("tests/user_journeys/test_data/simple_environment.yaml"),
    ],
)
def test_user_login_and_create_shared_environment(
    base_url: str, specification_path: str
) -> None:
    """Test that a user can login and create an environment in a shared namespace."""
    api = utils.API(base_url=base_url)

    # Create a shared namespace; default permissions for namepace/environment
    # */* is admin
    namespace = api.create_namespace().json()["data"]["name"]

    dev_api = utils.API(
        base_url=base_url,
        token=api.create_token(
            namespace,
            "developer",
        ).json()[
            "data"
        ]["token"],
    )

    environment = dev_api.create_environment(
        namespace,
        specification_path,
    ).json()[
        "data"
    ]["specification"]["name"]

    api.delete_environment(namespace, environment)
    api.delete_namespace(namespace)


@pytest.mark.user_journey
def test_admin_set_active_build(base_url: str):
    """Test that an admin can delete environments."""
    specs = [
        "tests/user_journeys/test_data/simple_environment.yaml",
        "tests/user_journeys/test_data/simple_environment2.yaml",
    ]
    api = utils.API(base_url=base_url)
    namespace = api.create_namespace().json()["data"]["name"]
    envs = set()
    for spec in specs:
        envs.add(
            api.create_environment(namespace, spec).json()["data"]["specification"][
                "name"
            ]
        )

    environment = list(envs)[0]
    builds = api.get_builds(
        namespace=namespace,
        environment=environment,
    )

    # The environment in the two specs has the same name, so there should be
    # two builds for this environment in this namespace
    assert len(builds) == 2

    build_ids = [build["id"] for build in builds]

    assert api.get_environment(
        namespace=namespace,
        environment=environment,
    )[
        "current_build_id"
    ] == max(build_ids)

    api.set_active_build(
        namespace=namespace, environment=environment, build_id=min(build_ids)
    )

    assert api.get_environment(
        namespace=namespace,
        environment=environment,
    )[
        "current_build_id"
    ] == min(build_ids)

    for env in envs:
        api.delete_environment(namespace, env)
    api.delete_namespace(namespace)


@pytest.mark.user_journey
def test_failed_build_logs(base_url: str):
    """Test that a user can access logs for a failed build."""
    api = utils.API(base_url=base_url)
    namespace = "default"
    build_request = api.create_environment(
        namespace,
        "tests/user_journeys/test_data/broken_environment.yaml",
    ).json()

    assert build_request["data"]["status"] == "FAILED"
    assert (
        "invalidpackagenamefaasdfagksdjfhgaskdf"
        in api.get_logs(build_request["data"]["id"]).text
    )

    api.delete_environment(
        namespace,
        build_request["data"]["specification"]["name"],
    )


@pytest.mark.user_journey
def test_cancel_build(base_url: str):
    """Test that a user cancel a build in progress."""
    api = utils.API(base_url=base_url)
    namespace = "default"
    build_id = api.create_environment(
        namespace,
        "tests/user_journeys/test_data/complicated_environment.yaml",
        wait=False,
    ).json()["data"]["build_id"]

    assert api.get_build_status(build_id) in [
        utils.BuildStatus.QUEUED,
        utils.BuildStatus.BUILDING,
    ]
    api.cancel_build(build_id)

    def check_status():
        status = api.get_build_status(build_id)
        if status in [utils.BuildStatus.QUEUED, utils.BuildStatus.BUILDING]:
            return False

        if status in [utils.BuildStatus.COMPLETED, utils.BuildStatus.FAILED]:
            raise ValueError(
                f"Build {build_id} {status.value.lower()}, but should have been canceled."
            )

        return status == utils.BuildStatus.CANCELED

    utils.wait_for_condition(check_status, timeout=60, interval=1)


@pytest.mark.user_journey
def test_get_lockfile(base_url: str):
    """Test that an admin can access a valid lockfile for a build."""
    api = utils.API(base_url=base_url)
    namespace = "default"
    build_request = api.create_environment(
        namespace,
        "tests/user_journeys/test_data/simple_environment.yaml",
    ).json()

    lockfile = api.get_lockfile(build_request["data"]["id"])

    packages = set(package["name"] for package in lockfile["package"])
    assert "python" in packages
    assert "fastapi" in packages

    api.delete_environment(namespace, build_request["data"]["specification"]["name"])



"""Interface between conda-store and conda

This module provides all the functionality that is required for
executing conda commands
"""

import bz2
import datetime
import json
import pathlib
import subprocess
import tempfile

import requests
import yaml
import yarl


def normalize_channel_name(channel_alias, channel):
    if channel.startswith("http"):
        return channel

    channel_alias = yarl.URL(channel_alias)
    return str(channel_alias / channel)


def conda_list(prefix, executable: str = "conda"):
    args = [executable, "list", "-p", prefix, "--json"]
    return json.loads(subprocess.check_output(args))


def conda_root_package_dir() -> pathlib.Path:
    args = ["conda", "info", "--json"]
    conf = json.loads(subprocess.check_output(args))
    prefix: pathlib.Path = pathlib.Path(conf["pkgs_dirs"][0])
    return prefix


def conda_pack(prefix, output, ignore_missing_files=True):
    from conda_pack import pack

    pack(
        prefix=str(prefix),
        output=str(output),
        ignore_missing_files=ignore_missing_files,
    )


def conda_lock(specification: "CondaSpecification", conda_exe: str = "mamba"):  # noqa
    from conda.models.dist import Dist
    from conda_lock.conda_lock import run_lock

    with tempfile.TemporaryDirectory() as tmpdir:
        environment_path = pathlib.Path(tmpdir) / "environment.yaml"
        lockfile_path = pathlib.Path(tmpdir) / "environment-lock.yaml"

        with environment_path.open("w") as f:
            f.write(specification.json())

        try:
            run_lock(
                environment_files=[environment_path],
                platforms=[conda_platform()],
                lockfile_path=lockfile_path,
                conda_exe=conda_exe,
            )
        except subprocess.CalledProcessError as e:
            raise ValueError(e.output)

        with lockfile_path.open() as f:
            lockfile = yaml.safe_load(f)

    conda_packages = []
    pip_packages = []

    for package in lockfile["package"]:
        if package["manager"] == "conda":
            dist = Dist.from_string(package["url"])
            conda_packages.append(
                {
                    "name": dist.name,
                    "build": dist.build,
                    "build_number": dist.build_number,
                    "constrains": None,
                    "depends": [],
                    "license": None,
                    "license_family": None,
                    "size": -1,
                    "subdir": dist.subdir,
                    "timestamp": None,
                    "version": dist.version,
                    "channel_id": dist.base_url,
                    "md5": package["hash"].get("md5"),
                    "sha256": package["hash"].get("sha256", ""),
                    "summary": None,
                    "description": None,
                }
            )
        elif package["manager"] == "pip":
            pip_packages.append(
                {
                    "name": package["name"],
                    "url": package["url"],
                    "version": package["version"],
                    "sha256": package["hash"]["sha256"],
                }
            )

    return {"conda": conda_packages, "pip": pip_packages}


def get_channel_url(channel: str) -> yarl.URL:
    # the conda main channel does not have a channeldata.json within
    # the https://conda.anaconda.org/<channel>/channeldata.json
    # so we replace the given url with it's equivalent alias
    channel_replacements = {
        "https://conda.anaconda.org/main/": yarl.URL(
            "https://repo.anaconda.com/pkgs/main"
        )
    }

    # Note: this doesn't use the / operator to append the trailing / character
    # because it's ignored on Windows. Instead, string substitution is used if
    # the / character is not present
    if channel.endswith("/"):
        normalized_channel_url = yarl.URL(channel)
    else:
        normalized_channel_url = yarl.URL(f"{channel}/")

    return channel_replacements.get(str(normalized_channel_url), yarl.URL(channel))


def download_repodata(
    channel: str,
    last_update: datetime.datetime = None,
    subdirs=None,
):
    """Download repodata for channel only if changed since last update

    A channel consists of several architectures: linux-32, linux-64,
    linux-aarch64, linux-armv6l, linux-armv7l, linux-ppc64le, noarch,
    osx-64, win-32, win-64, zos-z (possibly others).

    Check ``conda.base.constants.KNOWN_SUBDIRS`` for the full list.
    """
    subdirs = set(subdirs or [conda_platform(), "noarch"])

    channel_url = get_channel_url(channel)

    headers = {}
    if last_update:
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/If-Modified-Since
        # only download channeldata.json if it has been modified since
        # last update to minimize bandwidth usage of channel updates
        # most static web servers obey this header otherwise it will
        # be ignored
        headers["If-Modified-Since"] = last_update.strftime("%a, %d %b %Y %H:%M:%S GMT")

    response = requests.get(channel_url / "channeldata.json", headers=headers)
    if response.status_code == 304:  # 304 Not Modified since last_update
        return {"architectures": {}}
    response.raise_for_status()

    repodata = response.json()
    repodata["architectures"] = {}
    for subdir in subdirs:
        response = requests.get(
            channel_url / subdir / "repodata.json.bz2", headers=headers
        )
        if response.status_code == 304:  # 304 Not Modified since last_update
            continue
        response.raise_for_status()
        repodata["architectures"][subdir] = json.loads(bz2.decompress(response.content))
    return repodata


def conda_platform():
    """
    Return the platform (in ``conda`` style) the server is running on.

    It will be one of the values in ``conda.base.constants.KNOWN_SUBDIRS``.
    """
    from conda.base.context import context

    return context.subdir


def is_conda_prefix(conda_prefix: pathlib.Path):
    return (conda_prefix / "conda-meta/history").exists()



import json
import pathlib
import sys
import typing

from conda_store_server import action


@action.action
def action_install_lockfile(
    context,
    conda_lock_spec: typing.Dict,
    conda_prefix: pathlib.Path,
):
    lockfile_filename = pathlib.Path.cwd() / "conda-lock.yml"
    with lockfile_filename.open("w") as f:
        json.dump(conda_lock_spec, f)

    command = [
        sys.executable,
        "-m",
        "conda_lock",
        "install",
        "--validate-platform",
        "--log-level",
        "INFO",
        "--prefix",
        str(conda_prefix),
        str(lockfile_filename),
    ]

    context.run_command(command, check=True)
